import numpy as np
from cebra import CEBRA

from amadeusgpt.programs.api_registry import register_integration_api


# those functions need to have a self keyword to access attributes in AnimalBehaviorAnalysis
@register_integration_api
def get_cebra_embedding(self, inputs: np.ndarray, n_dimension=3) -> np.ndarray:
    """
    This function takes non-centered keypoints and calculate the embeddings using the CEBRA algorithm.
    Parameters
    ----------
    inputs: np.ndarray 4d tensor of shape (n_frames, n_individuals, n_kpts, n_features)
    n_dimensions: int, optional, default to be 3
    Returns
    -------
    ndarray of shape (n_frames,  n_dimension)
    """
    features = inputs.reshape(inputs.shape[0], -1)
    features = np.nan_to_num(features)

    print("features shape", features.shape)
    cebra_params = dict(
        model_architecture="offset10-model",
        batch_size=512,
        learning_rate=3e-4,
        temperature=1.12,
        output_dimension=n_dimension,
        max_iterations=10,
        distance="cosine",
        conditional="time_delta",
        device="cpu",
        verbose=True,
        time_offsets=10,
    )
    print("got here1")
    model = CEBRA(**cebra_params)
    print("got here2")
    model.fit(features)
    print("got here3")
    embeddings = model.transform(features)
    print("got here4")
    return embeddings



from typing import Any, Dict, List

import matplotlib.path as mpath
import numpy as np
from numpy import ndarray
from scipy.spatial import ConvexHull
from amadeusgpt.analysis_objects.object import Object


class Animal(Object):
    def get_keypoint_names(self):
        """
        keypoint names should be the basic attributes
        """
        pass

    def summary(self):
        print(self.__class__.__name__)
        for attr_name in self.__dict__:
            print(f"{attr_name} has {self.__dict__[attr_name]}")


class AnimalSeq(Animal):
    """
    Because we support passing bodyparts indices for initializing an AnimalSeq object,
    body center, left, right, above, top are relative to the subset of keypoints.
    Attributes
    ----------
    self.wholebody: np.ndarray of keypoints of all bodyparts
    self.keypoint
    """

    def __init__(self, animal_name: str, keypoints: ndarray, keypoint_names: List[str]):
        self.name = animal_name
        self.whole_body: ndarray = keypoints
        self.keypoint_names = keypoint_names
        self._paths = []
        self.state = {}
        self.kinematics_types = ["speed", "acceleration"]
        self.bodypart_relation = ["bodypart_pairwise_distance"]
        self.support_body_orientation = False
        self.support_head_orientation = False
        # self.keypoints are updated by indices of keypoint names given
        keypoint_indices = [keypoint_names.index(name) for name in keypoint_names]
        self.keypoints = self.whole_body[:, keypoint_indices]
        self.center = np.nanmedian(self.whole_body, axis=1)

    def update_roi_keypoint_by_names(self, keypoint_names: List[str]):
        # update self.keypoints based on keypoint names given
        if keypoint_names is None:
            return
        keypoint_indices = [self.keypoint_names.index(name) for name in keypoint_names]
        self.keypoints = self.whole_body[:, keypoint_indices]

    def restore_roi_keypoint(self):
        self.keypoints = self.whole_body

    def set_body_orientation_keypoints(
        self, body_orientation_keypoints: Dict[str, Any]
    ):
        self.neck_name = body_orientation_keypoints["neck"]
        self.tail_base_name = body_orientation_keypoints["tail_base"]
        self.animal_center_name = body_orientation_keypoints["animal_center"]
        self.support_body_orientation = True

    def set_head_orientation_keypoints(
        self, head_orientation_keypoints: Dict[str, Any]
    ):
        self.nose_name = head_orientation_keypoints["nose"]
        self.neck_name = head_orientation_keypoints["neck"]
        self.support_head_orientation = True

    # all the properties cannot be cached because update could happen
    def get_paths(self):
        paths = []
        for ind in range(self.whole_body.shape[0]):
            paths.append(self.get_path(ind))
        return paths

    def get_path(self, ind):
        xy = self.whole_body[ind]
        xy = np.nan_to_num(xy)
        if np.all(xy == 0):
            return None

        hull = ConvexHull(xy)
        vertices = hull.vertices
        path_data = []
        path_data.append((mpath.Path.MOVETO, xy[vertices[0]]))
        for point in xy[vertices[1:]]:
            path_data.append((mpath.Path.LINETO, point))
        path_data.append((mpath.Path.CLOSEPOLY, xy[vertices[0]]))
        codes, verts = zip(*path_data)
        return mpath.Path(verts, codes)

    def get_keypoints(self, average_keypoints=False) -> ndarray:
        assert (
            len(self.keypoints.shape) == 3
        ), f"keypoints shape is {self.keypoints.shape}"
        if not average_keypoints:
            return self.keypoints
        else:
            return np.nanmean(self.keypoints, axis=1)

    def get_center(self):
        """
        median is more robust than mean
        """
        return np.nanmedian(self.keypoints, axis=1).squeeze()

    def get_xmin(self):
        return np.nanmin(self.keypoints[..., 0], axis=1)

    def get_xmax(self):
        return np.nanmax(self.keypoints[..., 0], axis=1)

    def get_ymin(self):
        return np.nanmin(self.keypoints[..., 1], axis=1)

    def get_ymax(self):
        return np.nanmax(self.keypoints[..., 1], axis=1)

    def get_zmin(self):
        return np.nanmin(self.keypoints[..., 2], axis=1)

    def get_zmax(self):
        return np.nanmax(self.keypoints[..., 2], axis=1)

    def get_keypoint_names(self):
        return self.keypoint_names
    

    def query_states(self, query: str) -> ndarray:
        assert query in [
            "speed",
            "acceleration_mag",
            "bodypart_pairwise_distance",
        ], f"{query} is not supported"

        if query == "speed":
            self.state[query] = self.get_speed()
        elif query == "acceleration_mag":
            self.state[query] = self.get_acceleration_mag()
        elif query == "bodypart_pairwise_distance":
            self.state[query] = self.get_bodypart_wise_relation()

        return self.state[query]

    def get_velocity(self) -> ndarray:
        keypoints = self.get_keypoints()
        velocity = np.diff(keypoints, axis=0) / 30
        velocity = np.concatenate([np.zeros((1,) + velocity.shape[1:]), velocity])
        assert len(velocity.shape) == 3
        return velocity

    def get_speed(self) -> ndarray:
        keypoints = self.get_keypoints()
        velocity = (
            np.diff(keypoints, axis=0) / 30
        )  # divided by frame rate to get speed in pixels/second
        # Pad velocities to match the original shape
        velocity = np.concatenate([np.zeros((1,) + velocity.shape[1:]), velocity])
        # Compute the speed from the velocity
        speed = np.linalg.norm(velocity, axis=-1)
        speed = np.expand_dims(speed, axis=-1)
        assert len(speed.shape) == 3
        return speed

    def get_acceleration(self) -> ndarray:
        velocities = self.get_velocity()
        accelerations = (
            np.diff(velocities, axis=0) / 30
        )  # divided by frame rate to get acceleration in pixels/second^2
        # Pad accelerations to match the original shape
        accelerations = np.concatenate(
            [np.zeros((1,) + accelerations.shape[1:]), accelerations], axis=0
        )
        assert len(accelerations.shape) == 3
        return accelerations

    def get_acceleration_mag(self) -> ndarray:
        """
        Returns the magnitude of the acceleration vector
        """
        accelerations = self.get_acceleration()
        acceleration_mag = np.linalg.norm(accelerations, axis=-1)
        acceleration_mag = np.expand_dims(acceleration_mag, axis=-1)
        assert len(acceleration_mag.shape) == 3
        return acceleration_mag

    def get_bodypart_wise_relation(self):
        keypoints = self.get_keypoints()
        diff = keypoints[..., np.newaxis, :, :] - keypoints[..., :, np.newaxis, :]
        sq_dist = np.sum(diff**2, axis=-1)
        distances = np.sqrt(sq_dist)
        return distances

    def get_body_cs(
        self,
    ):
        # this only works for topview
        neck_index = self.keypoint_names.index(self.neck_name)
        tailbase_index = self.keypoint_names.index(self.tail_base_name)
        neck = self.whole_body[:, neck_index]
        tailbase = self.whole_body[:, tailbase_index]
        body_axis = neck - tailbase
        body_axis_norm = body_axis / np.linalg.norm(body_axis, axis=1, keepdims=True)
        # Get a normal vector pointing left
        mediolat_axis_norm = body_axis_norm[:, [1, 0]].copy()
        mediolat_axis_norm[:, 0] *= -1
        nrows = len(body_axis_norm)
        animal_cs = np.zeros((nrows, 3, 3))
        rot = np.stack((body_axis_norm, mediolat_axis_norm), axis=2)
        animal_cs[:, :2, :2] = rot
        animal_center_index = self.keypoint_names.index(self.animal_center_name)
        animal_cs[:, :, 2] = np.c_[
            self.whole_body[:, animal_center_index], np.ones(nrows)
        ]  # center back

        return animal_cs

    def calc_head_cs(self):
        nose_index = self.keypoint_names.index(self.nose_name)
        nose = self.whole_body[:, nose_index]
        neck_index = self.keypoint_names.index(self.neck_name)
        neck = self.whole_body[:, neck_index]
        head_axis = nose - neck
        head_axis_norm = head_axis / np.linalg.norm(head_axis, axis=1, keepdims=True)
        # Get a normal vector pointing left
        mediolat_axis_norm = head_axis_norm[:, [1, 0]].copy()
        mediolat_axis_norm[:, 0] *= -1
        nrows = len(head_axis_norm)
        mouse_cs = np.zeros((nrows, 3, 3))
        rot = np.stack((head_axis_norm, mediolat_axis_norm), axis=2)
        mouse_cs[:, :2, :2] = rot
        mouse_cs[:, :, 2] = np.c_[neck, np.ones(nrows)]
        return mouse_cs


if __name__ == "__main__":
    # unit testing the shape of kinematics data
    # acceleration, acceleration_mag, velocity, speed, and keypoints

    from amadeusgpt.config import Config
    from amadeusgpt.main import AMADEUS

    config = Config(
        "/Users/shaokaiye/AmadeusGPT-dev/amadeusgpt/configs/MausHaus_template.yaml"
    )
    amadeus = AMADEUS(config)
    analysis = amadeus.get_analysis()
    # get an instance of animal
    animal = analysis.animal_manager.get_animals()[0]

    print("velocity shape", animal.get_velocity().shape)
    print("speed shape", animal.get_speed().shape)
    print("acceleration shape", animal.get_acceleration().shape)
    print("acceleration_mag shape", animal.get_acceleration_mag().shape)

    print(animal.query_states("acceleration_mag").shape)



import types

from amadeusgpt.behavior_analysis.identifier import Identifier
from amadeusgpt.managers import (AnimalManager, EventManager, GUIManager,
                                 Manager, ObjectManager, RelationshipManager,
                                 VisualManager)
from amadeusgpt.programs.api_registry import (DEFAULT_REGISTRY,
                                              INTEGRATION_API_REGISTRY)


class AnimalBehaviorAnalysis:
    """
    This class holds methods and objects that are useful for analyzing animal behavior.
    It owns multiple manager classes that are responsible for different aspects of the analysis.
    """

    def __init__(self, identifier: Identifier, **kwargs):

        # animal manager needs keypoint_file_path and model_manager for pose
        self.animal_manager = AnimalManager(identifier)

        # object manager needs sam_info, seriralized pickle objects
        self.object_manager = ObjectManager(identifier, self.animal_manager)

        # relationship manager needs animal_manager and object_manager
        self.relationship_manager = RelationshipManager(
            identifier, self.animal_manager, self.object_manager
        )

        # event manager needs reference to object_manager, animal_manager, and relationship_manager
        self.event_manager = EventManager(
            identifier,
            self.object_manager,
            self.animal_manager,
            self.relationship_manager,
        )

        # some managers need references to others to do their job
        self.visual_manager = VisualManager(
            identifier, self.animal_manager, self.object_manager
        )

        self.gui_manager = GUIManager(identifier, self.object_manager)

        # check all attributes that are inheritance of manager classes
        # and attach them as methods to the main class
        self._attach_manager_methods()

        # attach the integration methods to the main class
        self._attach_integration_methods()

    # Put the methods of the managers in the main behavior analysis class
    # So we can save tokens and make it easier for LLM to learn to use the methods
    def _attach_manager_methods(self):
        method_names = set()
        for attr in dir(self):
            if isinstance(getattr(self, attr), Manager):
                manager = getattr(self, attr)
                cls_name = manager.__class__.__name__
                for method_name in DEFAULT_REGISTRY[cls_name]:
                    if hasattr(manager, method_name):

                        method = getattr(manager, method_name)
                        setattr(
                            self,
                            method_name,
                            method.__get__(manager, manager.__class__),
                        )
                        if method_name not in method_names:
                            method_names.add(method_name)
                        else:
                            raise ValueError(
                                f"Method {method_name} already exists in the class"
                            )

    def _attach_integration_methods(self):
        ### adding the methods from the dictionary to the class
        for method_name, func_info in INTEGRATION_API_REGISTRY.items():
            # this bound still needs to happen before we can register the function into a instance method at this class
            bound_method = types.MethodType(func_info["func"], self)
            setattr(self, method_name, bound_method)

    def summary(self, manager_name=None):
        for manager in [
            self.animal_manager,
            self.object_manager,
            self.relationship_manager,
            self.event_manager,
            self.visual_manager,
        ]:
            if manager_name is not None:
                if manager.__class__.__name__ != manager_name:
                    continue
            print(manager.__class__.__name__)
            manager.summary()


if __name__ == "__main__":

    pass



from enum import IntEnum
from typing import Any, Dict, List, Union

import numpy as np
from numpy import ndarray
from scipy.spatial.distance import cdist

from .animal import AnimalSeq
from .base import AnalysisObject
from .object import Object


class Orientation(IntEnum):
    FRONT = 1
    BACK = 2
    LEFT = 3
    RIGHT = 4


def calc_orientation_in_egocentric_animal(animal_seq, p):
    "Express the 2D points p into the mouse-centric coordinate system."
    mouse_cs = animal_seq.get_body_cs()
    mouse_cs_inv = np.full_like(mouse_cs, np.nan)
    valid = np.isclose(np.linalg.det(mouse_cs[:, :2, :2]), 1)
    mouse_cs_inv[valid] = np.linalg.inv(mouse_cs[valid])
    if p.ndim == 2:
        p = np.pad(p, pad_width=((0, 0), (0, 1)), mode="constant", constant_values=1)
        p_in_mouse = np.squeeze(mouse_cs_inv @ p[:, :, None])
    else:
        p_in_mouse = mouse_cs_inv @ [*p, 1]  # object center in mouse coordinate system
    p_in_mouse = p_in_mouse[:, :2]
    theta = np.arctan2(
        p_in_mouse[:, 1], p_in_mouse[:, 0]
    )  # relative angle between the object and the mouse body axis
    theta = np.rad2deg(theta % (2 * np.pi))
    orientation = np.zeros(theta.shape[0])
    np.place(orientation, np.logical_or(theta >= 330, theta <= 30), Orientation.FRONT)
    np.place(orientation, np.logical_and(theta >= 150, theta <= 210), Orientation.BACK)
    np.place(orientation, np.logical_and(theta > 30, theta < 150), Orientation.RIGHT)
    np.place(orientation, np.logical_and(theta > 210, theta < 330), Orientation.LEFT)
    return orientation


def calc_angle_between_2d_coordinate_systems(cs1, cs2):
    R1 = cs1[:, :2, :2]
    R2 = cs2[:, :2, :2]
    dot = np.einsum("ij, ij -> i", R1[:, 0], R2[:, 0])
    return np.rad2deg(np.arccos(dot))


def get_pairwise_distance(arr1: np.ndarray, arr2: np.ndarray):
    # we want to make sure this uses a fast implementation
    # arr: (n_frame,  n_kpts, 2)
    assert len(arr1.shape) == 3 and len(arr2.shape) == 3
    # pariwise distance (n_frames, n_kpts, n_kpts)
    pairwise_distances = np.ones((arr1.shape[0], arr1.shape[1], arr2.shape[1])) * 100000
    for frame_id in range(arr1.shape[0]):
        # should we use the mean of all keypoints for the distance?
        pairwise_distances[frame_id] = cdist(arr1[frame_id], arr2[frame_id])

    return pairwise_distances


def calc_angle_in_egocentric_animal(mouse_cs_inv, p):
    "Express the 2D points p into the mouse-centric coordinate system."
    if p.ndim == 2:
        p = np.pad(p, pad_width=((0, 0), (0, 1)), mode="constant", constant_values=1)
        p_in_mouse = np.squeeze(mouse_cs_inv @ p[:, :, None])
    else:
        p_in_mouse = mouse_cs_inv @ [*p, 1]  # object center in mouse coordinate system
    p_in_mouse = p_in_mouse[:, :2]
    theta = np.arctan2(
        p_in_mouse[:, 1], p_in_mouse[:, 0]
    )  # relative angle between the object and the mouse body axis
    theta = np.rad2deg(theta % (2 * np.pi))

    return theta


class Relationship(AnalysisObject):
    """
    We never want to cache this.
    data: This attribute is a dictionary that contains the relationship between two objects
    each value of self.data is a numpy array that is either boolean or a float
    """

    def __getitem__(self, key):
        return self.data[key]

    def __setitem__(self, key, value):
        self.data[key] = value

    def get_name(self):
        return self.__name__

    def query_relationship(self, query_name: str) -> ndarray:
        ret = self.data[query_name]
        return ret

    def summary(self):
        print(self.__class__.__name__)
        for attr_name in self.__dict__:
            print(f"{attr_name} has {self.__dict__[attr_name]}")


class AnimalObjectRelationship(Relationship):
    """
    To be referenced in the animal class
    """

    def __init__(
        self,
        animal: AnimalSeq,
        other_obj: Object,
        sender_animal_bodyparts_names: Union[List[str], None] = None,
    ):
        self.sender_animal_name = animal.get_name()
        self.object_name = other_obj.get_name()
        self.sender_animal_bodyparts_names = sender_animal_bodyparts_names

        self.data: Dict[str, Any] = self._relationships_with_static_object(
            animal, other_obj
        )

    def __eq__(self, other: Relationship) -> bool:
        return (
            self.sender_animal_name == other.sender_animal_name
            and self.object_name == other.object_name
            and self.sender_animal_bodyparts_names
            == other.sender_animal_bodyparts_names
        )

    def _relationships_with_static_object(
        self, animal: AnimalSeq, other_obj: Object
    ) -> Dict[str, Any]:
        if self.sender_animal_bodyparts_names is not None:
            # the keypoints of animal get updated when we update the roi bodypart names
            animal.update_roi_keypoint_by_names(self.sender_animal_bodyparts_names)

        c = other_obj.get_center()

        distance = np.linalg.norm(animal.get_center() - c, axis=1)
        overlap = other_obj.Path.contains_points(animal.get_center())
        orientation = None
        if animal.support_body_orientation:
            orientation = calc_orientation_in_egocentric_animal(
                animal, other_obj.get_center()
            )

        ret = {
            "distance": distance,
            "overlap": overlap,
        }

        if orientation is not None:
            ret["orientation"] = orientation
        return ret


class AnimalAnimalRelationship(Relationship):
    """
    To be referenced in the animal class
    """

    def __init__(
        self,
        animal: AnimalSeq,
        other_animal: AnimalSeq,
        sender_animal_bodyparts_names: Union[List[str], None] = None,
        receiver_animal_bodyparts_names: Union[List[str], None] = None,
    ):

        self.sender_animal_name = animal.get_name()
        self.receiver_animal_name = other_animal.get_name()
        self.sender_animal_bodyparts_names = sender_animal_bodyparts_names
        self.receiver_animal_bodyparts_names = receiver_animal_bodyparts_names
        self.object_name = None
        self.data: Dict[str, Any] = self._animal_animal_relationship(
            animal, other_animal
        )

    def __eq__(self, other: Relationship) -> bool:
        return (
            self.sender_animal_name == other.sender_animal_name
            and self.receiver_animal_name == other.receiver_animal_name
            and self.sender_animal_bodyparts_names
            == other.sender_animal_bodyparts_names
            and self.receiver_animal_bodyparts_names
            == other.receiver_animal_bodyparts_names
        )

    def _animal_animal_relationship(
        self,
        sender_animal: AnimalSeq,
        receiver_animal: AnimalSeq,
    ) -> Dict[str, Any]:
        """
        This is the relationship between two animals
        """
        if self.sender_animal_bodyparts_names is not None:
            sender_animal.update_roi_keypoint_by_names(
                self.sender_animal_bodyparts_names
            )
        if self.receiver_animal_bodyparts_names is not None:
            receiver_animal.update_roi_keypoint_by_names(
                self.receiver_animal_bodyparts_names
            )

        # other animal is to the left of this animal
        distance = np.linalg.norm(
            sender_animal.get_center() - receiver_animal.get_center(), axis=1
        )

        # I have _coords for both this and other because people could want a subset of animal keypoints
        overlap = []
        # we only do nan to num here because doing it in other places give bad looking trajectory
        robust_center = np.nan_to_num(sender_animal.get_center())
        for path_id, other_path in enumerate(receiver_animal.get_paths()):
            if other_path is None:
                overlap.append(False)
                continue
            overlap.append(other_path.contains_point(np.array(robust_center[path_id])))
        overlap = np.array(overlap)

        angles = None
        head_angles = None
        orientation = None

        if sender_animal.support_body_orientation:
            angles = calc_angle_between_2d_coordinate_systems(
                sender_animal.get_body_cs(), receiver_animal.get_body_cs()
            )
            orientation = calc_orientation_in_egocentric_animal(
                sender_animal, receiver_animal.get_center()
            )

        if sender_animal.support_head_orientation:
            mouse_cs = sender_animal.calc_head_cs()
            head_cs_inv = []
            mouse_cs_inv = np.full_like(mouse_cs, np.nan)
            valid = np.isclose(np.linalg.det(mouse_cs[:, :2, :2]), 1)
            mouse_cs_inv[valid] = np.linalg.inv(mouse_cs[valid])
            head_cs_inv.append(mouse_cs_inv)
            head_angles = calc_angle_in_egocentric_animal(
                head_cs_inv, receiver_animal.get_center()
            )

        # relative_velocity = (
        #     sender_animal.get_velocity() - receiver_animal.get_velocity()
        # )
        # relative_velocity_magnitude = np.linalg.norm(relative_velocity, axis=2)
        # # Then, average these magnitudes over all keypoints for each frame
        # relative_speed = np.nanmean(relative_velocity_magnitude, axis=1)

        sender_pos = sender_animal.get_center()
        receiver_pos = receiver_animal.get_center()
        direction_vector = receiver_pos - sender_pos
        sender_velocity = np.nanmean(sender_animal.get_velocity(), axis=1)
        norm_direction_vector = direction_vector / np.linalg.norm(direction_vector)
        relative_speed = np.einsum("ij,ij->i", sender_velocity, norm_direction_vector)

        closest_distance = np.nanmin(
            get_pairwise_distance(sender_animal.keypoints, receiver_animal.keypoints),
            axis=(1, 2),
        )
        ret = {
            "distance": distance,
            "overlap": overlap,
            "closest_distance": closest_distance,
            "relative_speed": relative_speed,
        }

        if head_angles is not None:
            ret["relative_head_angle"] = head_angles

        if angles is not None:
            ret["relative_angle"] = angles
        if orientation is not None:
            ret["orientation"] = orientation

        return ret



import warnings

from amadeusgpt.config import Config
from amadeusgpt.programs.sandbox import Sandbox
##########
# all these are providing the customized classes for the code execution
##########
from amadeusgpt.utils import *

warnings.filterwarnings("ignore")
import glob
import os
import pickle
from pathlib import Path

from amadeusgpt.analysis_objects.llm import (CodeGenerationLLM, SelfDebugLLM,
                                             VisualLLM)
from amadeusgpt.behavior_analysis.identifier import Identifier
from amadeusgpt.integration_module_hub import IntegrationModuleHub
from amadeusgpt.programs.task_program_registry import TaskProgramLibrary


class AMADEUS:
    def __init__(self, config: Config | dict, use_vlm=False, movie2keypoint_func=None):
        self.config = config
        ### fields that decide the behavior of the application
        self.use_self_debug = True
        self.use_diagnosis = False
        self.use_behavior_modules_in_context = True
        self.smart_loading = False
        self.load_module_top_k = 3
        self.module_threshold = 0.3
        ### fields that serve as important storage
        # for long-term memory
        self.integration_module_hub = IntegrationModuleHub()

        ### For the sake of multiple animal, we store multiple sandboxes
        ### the example {video_file_path : sandbox }
        self.result_folder: str = config["data_info"]["result_folder"]
        self.data_folder = config["data_info"]["data_folder"]
        self.video_suffix = config["data_info"]["video_suffix"]

        self.video_file_paths, self.keypoint_file_paths = (
            self.fetch_data_from_data_folder(movie2keypoint_func)
        )

        self.sandbox = Sandbox(config, self.video_file_paths, self.keypoint_file_paths)

        self.code_generator_llm = CodeGenerationLLM(config)
        self.self_debug_llm = SelfDebugLLM(config)
        self.visual_llm = VisualLLM(config)

        ####

        ## register the llm to the sandbox

        self.sandbox.register_llm("code_generator", self.code_generator_llm)
        self.sandbox.register_llm("visual_llm", self.visual_llm)
        if self.use_self_debug:
            self.sandbox.register_llm("self_debug", self.self_debug_llm)

        # can only do this after the register process
        if use_vlm:
            self.sandbox.configure_using_vlm()

    def fetch_data_from_data_folder(self, movie2keypoint_func):
        """
        1) video file exists, keypoint file does not exist
        2) video file and keypoint file both exist and one-to-one mapping
        3) many video files and one keypoint file (many-to-one mapping), perhaps for 3D data
        4) none of video file and keypoint file exist
        5) only keypoint file exists

        Returns
        -------
        video_file_paths: list[str]
            a list of video file paths
        keypoint_file_paths: list[str]
            a list of keypoint file paths

        """
        video_files = glob.glob(os.path.join(self.data_folder, f"*{self.video_suffix}"))
        video_files = [
            video_file for video_file in video_files if "labeled" not in video_file
        ]
        keypoint_files = glob.glob(os.path.join(self.data_folder, "*.h5"))
        # case 1
        if movie2keypoint_func is None:
            if len(video_files) > 0 and len(keypoint_files) == 0:
                return video_files, [""] * len(video_files)
            # case 2
            elif len(video_files) > 0 and len(video_files) == len(keypoint_files):
                return video_files, self.get_DLC_keypoint_files(video_files)
            # case 3
            elif len(video_files) > 1 and len(keypoint_files) == 1:
                print(
                    "We assume this is 3D data with multiple video files and one keypoint file"
                )
                return video_files, [keypoint_files[0]] * len(video_files)
            # case 4
            elif len(video_files) == 0 and len(keypoint_files) == 0:
                raise ValueError(
                    "No video files and keypoint files found in the data folder"
                )
            # case 5
            elif len(video_files) == 0 and len(keypoint_files) > 0:
                print("No video found. We proceed with the keypoint file only")
                return [""] * len(keypoint_files), keypoint_files
            else:
                return video_files, keypoint_files
        else:
            if len(video_files) > 0:
                return video_files, [
                    movie2keypoint_func(video_file) for video_file in video_files
                ]
            else:
                raise ValueError("No video files found in the data folder")

    def get_DLC_keypoint_files(self, video_file_paths: list[str]):
        ret = []
        # how to get the filename from the path file
        video_folder = Path(video_file_paths[0]).parent
        for video_file_path in video_file_paths:
            videoname = Path(video_file_path).stem
            if len(glob.glob(str(video_folder / f"{videoname}*.h5"))) > 0:
                keypoint_file_path = glob.glob(str(video_folder / f"{videoname}*.h5"))[
                    0
                ]
            else:
                keypoint_file_path = ""
            ret.append(keypoint_file_path)

        return ret

    def match_integration_module(self, user_query: str) -> list:
        """
        Return a list of matched integration modules
        """
        sorted_query_results = self.integration_module_hub.match_module(user_query)
        if len(sorted_query_results) == 0:
            return None
        modules = []
        for i in range(min(self.load_module_top_k, len(sorted_query_results))):
            query_result = sorted_query_results[i]
            query_module = query_result[0]
            query_score = query_result[1][0][0]
            if query_score > self.module_threshold:
                modules.append(query_module)

                # parse the query result by loading active loading
        return modules

    def step(self, user_query: str) -> QA_Message:
        integration_module_names = self.match_integration_module(user_query)

        self.sandbox.update_matched_integration_modules(integration_module_names)
        qa_message = self.sandbox.llm_step(user_query)

        return qa_message

    def get_video_file_paths(self) -> list[str]:
        return self.video_file_paths

    def get_keypoint_file_paths(self) -> list[str]:
        return self.keypoint_file_paths

    def get_behavior_analysis(
        self, video_file_path: str = "", keypoint_file_path: str = ""
    ):
        """
        Every sandbox stores a unique "behavior analysis" instance in its namespace
        Therefore, get analysis gets the current sandbox's analysis.
        """
        identifier = Identifier(self.config, video_file_path, keypoint_file_path)
        analysis = self.sandbox.namespace_dict[identifier]["behavior_analysis"]

        return analysis

    def run_task_program(self, task_program_name: str):
        """
        Execute the task program on the currently holding sandbox
        Parameters
        -----------
        config: a config specifies the movie file and the keypoint file to run task program
        task_program_name: the name of the task program to run

        """
        return self.sandbox.run_task_program(task_program_name)

    def register_task_program(self, task_program, creator="human"):
        TaskProgramLibrary.register_task_program(creator=creator)(task_program)

    def get_messages(self):
        return self.sandbox.message_cache

    def get_task_programs(self):
        return TaskProgramLibrary.get_task_programs()


if __name__ == "__main__":
    from amadeusgpt.analysis_objects.llm import VisualLLM
    from amadeusgpt.config import Config

    config = Config("amadeusgpt/configs/EPM_template.yaml")
    amadeus = AMADEUS(config)



import base64
import io
import json
import os
import re
import time
import traceback

import cv2
import numpy as np
import openai
from openai import OpenAI

from amadeusgpt.programs.sandbox import Sandbox
from amadeusgpt.utils import AmadeusLogger, QA_Message, create_qa_message

from .base import AnalysisObject


class LLM(AnalysisObject):
    total_tokens = 0
    prices = {
        "gpt-4o": {"input": 5 / 10**6, "output": 15 / 10**6},
        "gpt-4o-mini": {"input": 0.15 / 10**6, "output": 0.6 / 10**6},
    }
    total_cost = 0

    def __init__(self, config):
        self.config = config

        self.max_tokens = config["llm_info"].get("max_tokens", 4096)
        self.gpt_model = config["llm_info"].get("gpt_model", "gpt-4o-mini")
        self.keep_last_n_messages = config.get("keep_last_n_messages", 2)

        # the list that is actually sent to gpt
        self.context_window = []
        # only for logging and long-term memory usage.
        self.history = []

    def encode_image(self, image_path):
        with open(image_path, "rb") as image_file:
            return base64.b64encode(image_file.read()).decode("utf-8")

    def speak(self):
        """
        Speak to the chat channel
        """
        raise NotImplementedError("This method should be implemented in the subclass")

    def parse_json_string(self, text):
        """
        Text contains a json string with json block
        """
        pattern = r"```json(.*?)```"
        json_string = re.findall(pattern, text, re.DOTALL)[0]
        json_obj = json.loads(json_string)
        return json_obj

    def connect_gpt(self, messages, **kwargs):
        # if openai version is less than 1
        return self.connect_gpt_oai_1(messages, **kwargs)

    def connect_gpt_oai_1(self, messages, **kwargs):
        """
        This is routed to openai > 1.0 interfaces
        """

        if self.config.get("use_streamlit", False):
            if "OPENAI_API_KEY" in os.environ:
                openai.api_key = os.environ["OPENAI_API_KEY"]
        else:
            openai.api_key = os.environ["OPENAI_API_KEY"]
        response = None
        # gpt_model is default to be the cls.gpt_model, which can be easily set
        gpt_model = self.gpt_model
        # in streamlit app, "gpt_model" is set by the text box

        client = OpenAI()

        if self.config.get("use_streamlit", False):
            if "gpt_model" in st.session_state:
                self.gpt_model = st.session_state["gpt_model"]

        configurable_params = ["gpt_model", "max_tokens", "temperature"]

        for param in configurable_params:
            if param in kwargs:
                setattr(self, param, kwargs[param])

        # the usage was recorded from the last run. However, since we have many LLMs that
        # share the call of this function, we will need to store usage and retrieve them from the database class
        num_retries = 3
        for _ in range(num_retries):
            try:
                json_data = {
                    "model": self.gpt_model,
                    "messages": messages,
                    "max_tokens": self.max_tokens,
                    "temperature": 0.0,
                }
                response = client.chat.completions.create(**json_data)

                LLM.total_tokens = (
                    LLM.total_tokens
                    + response.usage.prompt_tokens
                    + response.usage.completion_tokens
                )
                LLM.total_cost += (
                    LLM.prices[self.gpt_model]["input"] * response.usage.prompt_tokens
                    + LLM.prices[self.gpt_model]["output"]
                    * response.usage.completion_tokens
                )

                print("current total cost", round(LLM.total_cost, 4), "$")
                print("current input tokens", response.usage.prompt_tokens)
                print("current accumulated tokens", LLM.total_tokens)
                # TODO we need to calculate the actual dollar cost
                break

            except Exception as e:

                error_message = traceback.format_exc()
                print("error", error_message)

                if "This model's maximum context" in error_message:
                    if len(self.context_window) > 2:
                        AmadeusLogger.info("doing active forgetting")
                        self.context_window.pop(1)
                        # and the corresponding bot answer
                        self.context_window.pop(1)

                elif "Rate limit reached" in error_message:
                    print("Hit rate limit. Sleeping for 10 sec")
                    time.sleep(10)

        return response

    # image list can be image byte or image array
    def prepare_multi_image_content(self, image_list):
        """ """
        encoded_image_list = []

        for image in image_list:
            # images from matplotlib etc.
            if isinstance(image, io.BytesIO):
                image_bytes = image
                base64_image = base64.b64encode(image_bytes.getvalue()).decode("utf-8")
            # images from opencv
            elif isinstance(image, np.ndarray):
                result, buffer = cv2.imencode(".jpeg", image)
                image_bytes = io.BytesIO(buffer)
                base64_image = base64.b64encode(image_bytes.getvalue()).decode("utf-8")

            encoded_image_list.append(base64_image)
        multi_image_content = [
            {
                "type": "image_url",
                "image_url": {"url": f"data:image/jpeg;base64,{encoded_image}"},
            }
            for encoded_image in encoded_image_list
        ]
        return multi_image_content

    def update_history(self, role, content, multi_image_content=None, in_place=False):
        """
        multi_image_content: can support multi image content for gpt-4o or euiqvalent
        in_place: always use the index 1 as the user message. This is for LLM does not need to keep a history
        """
        if role == "system":
            if len(self.history) > 0:
                self.history[0]["content"] = content
                self.context_window[0]["content"] = content
            else:
                self.history.append({"role": role, "content": content})
                self.context_window.append({"role": role, "content": content})
        else:

            if multi_image_content is None:
                new_message = {"role": role, "content": content}
            else:
                text_content = {"type": "text", "text": content}
                multi_image_content = [text_content] + multi_image_content
                new_message = {"role": role, "content": multi_image_content}

            self.history.append(new_message)
            # responsible for active forgetting
            num_AI_messages = (len(self.context_window) - 1) // 2
            if num_AI_messages == self.keep_last_n_messages:
                print("doing active forgetting")
                # we forget the oldest AI message and corresponding answer
                # index 0 is reserved for system prompt so we always pop index 1
                self.context_window.pop(1)
                self.context_window.pop(1)

            if in_place == True:
                assert (
                    len(self.context_window) <= 2
                ), "context window should have no more than 2 elements"
                if len(self.context_window) == 2:
                    self.context_window[1] = new_message
                else:
                    self.context_window.append(new_message)
            else:
                self.context_window.append(new_message)

    def clean_context_window(self):
        while len(self.context_window) > 1:
            AmadeusLogger.info("cleaning context window")
            self.context_window.pop()

    def parse_openai_response(cls, response):
        """
        Take the chat history, parse gpt's answer and execute the code

        Returns
        -------
        Dictionary
        """
        # there should be better way to do this
        if response is None:
            text = "Something went wrong"
        else:
            text = response.choices[0].message.content.strip()

        # we need to consider better ways to parse functions
        # and save them in a more structured way
        pattern = r"```python(.*?)```"
        function_code = re.findall(pattern, text, re.DOTALL)[0]

        # create a placeholder
        thought_process = text.replace(function_code, "<python_code>")

        return text, function_code, thought_process

    def get_system_prompt(self, sandbox):
        raise NotImplementedError("This method should be implemented in the subclass")


class VisualLLM(LLM):
    def __init__(self, config):
        super().__init__(config)

    def speak(self, sandbox: Sandbox, image: np.ndarray):
        """
        Only to comment about one image
        #1) What animal is there, how many and what superanimal model we should use
        #2) report the background object list
        #3) We format them in json format
        """
        from amadeusgpt.system_prompts.visual_llm import _get_system_prompt

        self.system_prompt = _get_system_prompt()
        multi_image_content = self.prepare_multi_image_content([image])

        self.update_history("system", self.system_prompt)
        self.update_history(
            "user",
            "here is the image",
            multi_image_content=multi_image_content,
            in_place=True,
        )
        response = self.connect_gpt(self.context_window, max_tokens=2000)
        text = response.choices[0].message.content.strip()

        print("description of the image frame provided")
        print(text)

        pattern = r"```json(.*?)```"
        if len(re.findall(pattern, text, re.DOTALL)) == 0:
            raise ValueError("can't parse the json string correctly", text)
        else:
            json_string = re.findall(pattern, text, re.DOTALL)[0]
            json_obj = json.loads(json_string)
            return json_obj


class CodeGenerationLLM(LLM):
    """
    Resource management for the behavior analysis part of the system
    """

    def __init__(self, config):
        super().__init__(config)

    def speak(
        self, sandbox: Sandbox, qa_message: QA_Message, share_video_file=True
    ) -> QA_Message:
        """ """
        query = qa_message.query

        from amadeusgpt.system_prompts.code_generator import _get_system_prompt

        core_api_docs = sandbox.get_core_api_docs()
        task_program_docs = sandbox.get_task_program_docs()

        if share_video_file:
            identifier = sandbox.identifiers[0]
        else:
            raise NotImplementedError("This is not implemented yet")

        behavior_analysis = sandbox.analysis_dict[identifier]
        scene_image = behavior_analysis.visual_manager.get_scene_image()
        keypoint_names = behavior_analysis.animal_manager.get_keypoint_names()
        object_names = behavior_analysis.object_manager.get_object_names()
        animal_names = behavior_analysis.animal_manager.get_animal_names()        
        use_3d = sandbox.config['keypoint_info'].get('use_3d', False)
        
        self.system_prompt = _get_system_prompt(
            core_api_docs,
            task_program_docs,
            scene_image,
            keypoint_names,
            object_names,
            animal_names,
            use_3d=use_3d,
        ) 

        self.update_history("system", self.system_prompt)

        self.update_history("user", query)

        response = self.connect_gpt(self.context_window, max_tokens=2000)
        text = response.choices[0].message.content.strip()
        # need to keep the memory of the answers from LLM
        self.update_history("assistant", text)

        function_code = None

        pattern = r"```python(.*?)```"
        if len(re.findall(pattern, text, re.DOTALL)) == 0:
            pass
        else:
            function_code = re.findall(pattern, text, re.DOTALL)[0]

        # it's a bit meaningless to copy this to every qa_message

        qa_message.code = function_code
        qa_message.chain_of_thought = text

        # this is for debug use
        with open("temp_answer.json", "w") as f:
            obj = {}
            obj["chain_of_thought"] = text
            obj["code"] = function_code
            obj["video_file_paths"] = sandbox.video_file_paths
            obj["keypoint_file_paths"] = sandbox.keypoint_file_paths
            if not isinstance(sandbox.config, dict):
                obj["config"] = sandbox.config.to_dict()
            else:
                obj["config"] = sandbox.config
            json.dump(obj, f, indent=4)

        return qa_message


class SelfDebugLLM(LLM):

    def __init__(self, config):
        super().__init__(config)

    def speak(self, qa_message):

        error_message = qa_message.error_message
        code = qa_message.code

        from amadeusgpt.system_prompts.self_debug import _get_system_prompt

        self.system_prompt = _get_system_prompt()
        self.update_history("system", self.system_prompt)
        print("the code that gave errors was", code)
        query = f""" The code that caused error was {code}
And the error message was {error_message}. 
All the modules were already imported so you don't need to import them again.
Can you correct the code? Make sure you only write one function which is the updated function.
"""
        self.update_history("user", query)
        response = self.connect_gpt(self.context_window, max_tokens=4096)
        text = response.choices[0].message.content.strip()
        print(text)
        pattern = r"```python(.*?)```"
        function_code = re.findall(pattern, text, re.DOTALL)[0]
        qa_message.code = function_code
        qa_message.chain_of_thought = text

        return qa_message

if __name__ == "__main__":
    from amadeusgpt.config import Config

    config = Config(
        "/Users/shaokaiye/AmadeusGPT-dev/amadeusgpt/configs/MausHaus_template.yaml"
    )



import os
import re
from typing import Any, Dict, List, Literal, Optional, Union

import numpy as np

from amadeusgpt.analysis_objects.event import Event, EventGraph
from amadeusgpt.analysis_objects.relationship import Relationship
from amadeusgpt.behavior_analysis.identifier import Identifier
from amadeusgpt.programs.api_registry import (register_class_methods,
                                              register_core_api)

from .animal_manager import AnimalManager
from .base import Manager, cache_decorator
from .object_manager import ObjectManager
from .relationship_manager import RelationshipManager


def find_complement_number(string):
    digits = ""
    for char in string:
        if char.isdigit() or char == ".":
            digits += char
    return str(360 - float(digits))


def find_complement_operator(string):
    operator = ""
    for char in string:
        if not char.isdigit() and not char == ".":
            operator += char
    operator = operator.strip()
    if operator == "<":
        return ">="
    elif operator == "<=":
        return ">"
    elif operator == ">":
        return "<="
    elif operator == ">=":
        return "<"


def process_animal_object_relation(
    relation_query: str, comparison: str, animal_object_relation: np.ndarray
):
    if relation_query in ["relatve_head_angle"]:
        complement_relation = find_complement_operator(comparison)
        digits = find_complement_number(comparison)
        complement_comparison = complement_relation + digits
        relation_string = "numeric_quantity" + complement_comparison
        complement_animal_object_relation = eval(relation_string)
        animal_object_relation |= complement_animal_object_relation
    return animal_object_relation


@register_class_methods
class EventManager(Manager):
    def __init__(
        self,
        identifier: Identifier,
        object_manager: ObjectManager,
        animal_manager: AnimalManager,
        relationship_manager: RelationshipManager,
        use_cache: bool = False,
    ):
        self.config = identifier.config
        super().__init__(self.config, use_cache=use_cache)

        self.object_manager = object_manager
        self.animal_manager = animal_manager
        self.relationship_manager = relationship_manager
        self.video_file_path = identifier.video_file_path
        if not os.path.exists(self.video_file_path):
            # no need to initialize
            return

        self.animals_object_events = []
        self.animals_animals_events = []
        self.animals_state_events = []

    @register_core_api
    def get_animals_object_events(
        self,
        object_name: Optional[str] = "",
        query: Optional[str] = "",
        negate: Optional[bool] = False,
        bodypart_names: Optional[Union[List[str], None]] = None,
        min_window: int = 0,
        max_window: int = 1000000,
        smooth_window_size: int = 3,
    ) -> List[Event]:
        """
        This function is used to retrieve events that involve the interactions between animals and objects.
        object_name : str
        This parameter represents the name of the object of interest. It is expected to be a string. Must be one of the object names returned from function get_object_names.
        query: str. Examples are 'overlap == True', 'distance==30', 'angle<20',
        bodypart_names: List[str], optional
           bodyparts of the animal. Default to be None which means all keypoints of the animal are considered.
        min_window: min length of the event to include
        max_window: max length of the event to include
        negate: bool, default false
           whether to negate the spatial events. For example, if negate is set True, inside roi would be outside roi
        """
        pattern = r"(==|<=|>=|<|>)"
        if bodypart_names is not None:
            self.animal_manager.update_roi_keypoint_by_names(bodypart_names)

        animals_objects_relations = (
            self.relationship_manager.get_animals_objects_relationships(
                animal_bodyparts_names=bodypart_names
            )
        )

        ret_events = []
        for animal_objects_relationship in animals_objects_relations:
            if animal_objects_relationship.object_name != object_name:
                continue
            # to construct the events
            mask: List[bool] = None
            sender_animal_name = animal_objects_relationship.sender_animal_name
            comparison_operator = re.findall(pattern, query)[0].strip()
            _query = query.split(comparison_operator)[0].strip()
            _comparison = comparison_operator + "".join(
                query.split(comparison_operator)[1:]
            )
            numeric_quantity = animal_objects_relationship.query_relationship(_query)
            relation_string = "numeric_quantity" + _comparison
            mask = eval(relation_string)
            mask = process_animal_object_relation(_query, _comparison, mask)

            events: List[Event] = Event.mask2events(
                mask,
                self.video_file_path,
                sender_animal_name,
                set(),
                {object_name},
                smooth_window_size=smooth_window_size,
            )

            if negate:
                events = Event.event_negate(events)

            events = Event.filter_events_by_duration(events, min_window, max_window)

            ret_events.extend(events)

        if bodypart_names is not None:
            self.animal_manager.restore_roi_keypoint()

        return ret_events

    # @cache_decorator
    @register_core_api
    def get_animals_state_events(
        self,
        mask: np.ndarray,
        min_window: int = 10,
        max_window: int = 1000000,
    ) -> List[Event]:
        """
        Parameters
        ----------
        mask: np.ndarray, optional.
            The mask must be of shape (n_frames, n_individuals). It is a boolean mask that describes the condition for the behavior.
            If n_individuals is 1, the shape should be (n_frames, 1)
        min_window : int, optional, default 10
            Only include events that are longer than min_window
        max_window : int, optional, default 1000000
            Only include events that are shorter than max_window
        Returns
        -------
        List[Event]
        --------
        """

        if len(mask.shape) == 1:
            mask = mask.reshape(-1, 1)

        ret_events = []       
        for animal_idx, sender_animal_name in enumerate(
            self.animal_manager.get_animal_names()
        ):
            # to construct the events

            events = Event.mask2events(
                mask[:, animal_idx],
                self.video_file_path,
                sender_animal_name,
                set(),
                set(),
            )       
            events = Event.filter_events_by_duration(events, min_window, max_window)        
            ret_events.extend(events)

        ret_events = sorted(ret_events, key=lambda x: x.start)

        return ret_events

    @cache_decorator
    def get_events_from_relationship(
        self,
        relationship: Relationship,
        relation_query: str,
        comparison: str,
        smooth_window_size: int,
    ) -> List[Event]:

        mask = relationship.query_relationship(relation_query)

        # determine whether the mask is a numpy of float or numpy of boolean

        if mask.dtype != bool:
            relation_string = "mask" + comparison
            mask = eval(relation_string)

        sender_animal_name = relationship.sender_animal_name
        receiver_animal_names = set([relationship.receiver_animal_name])
        if relationship.object_name is not None:
            object_names = set([relationship.object_name])
        else:
            object_names = set()

        events = Event.mask2events(
            mask,
            self.video_file_path,
            sender_animal_name,
            receiver_animal_names,
            object_names,
            smooth_window_size=smooth_window_size,
        )

        return events

    # @cache_decorator
    @register_core_api
    def get_animals_animals_events(
        self,
        cross_animal_query_list: List = [],
        bodypart_names: Optional[Union[List[str], None]] = None,
        otheranimal_bodypart_names: Optional[Union[List[str], None]] = None,
        min_window: int = 11,
        max_window: int = 1000000,
        smooth_window_size: int = 3,
    ) -> List[Event]:
        """
        The function is for capturing behaviors that involve multiple animals. Don't fill the bodypart_names and otheranimal_bodypart_names unless you know the names of the bodyparts.
        When multiple queries are passed, they are combined as logical_and, not logical_or or sequential.

        Parameters
        ----------
        cross_animal_query_list:
        list of queries describing conditions among animals. It takes form of {type_of_query}{operator}{value}
        Examples are:
        'orientation==Orientation.FRONT', 'orientation==Orientation.BACK', 'distance<10', 'relative_angle>30',
        The allowed type of query is chosen from ["overlap", "distance", "relative_speed", "orientation", "closest_distance", "relative_angle", "relative_head_angle"]
        bodypart_names:
        list of bodyparts for the this animal. By default, it is None, which means all bodyparts are included.
        If set, the given bodyparts will be the "ROI" bodyparts that are used for the queries.
        otheranimal_bodypart_names: list[str], optional
        list of bodyparts for the other animals. By default, it is None, which means all bodyparts are included.
        If set, the given bodyparts will be the "ROI" bodyparts that are used for the queries.
        min_window: int, optional, default 11
        Only include events that are longer than min_window
        max_window: int, optional, default 100000
        smooth_window_size: int, optional
        smooth window size for smoothing the events.
        Returns
        -------
        List[Event]
        Note
        ----
        To capture a range for a numerical query  (e.g., relative_speed) between 3 and 10, one can do:
        get_animals_animals_events(cross_animal_query_list = ['relative_speed<10', 'relative_speed>3'],

        """

        if min_window is None:
            min_window = 3
        if max_window is None:
            max_window = 1000000

        animals_animals_relationships = (
            self.relationship_manager.get_animals_animals_relationships(
                sender_animal_bodyparts_names=bodypart_names,
                receiver_animal_bodyparts_names=otheranimal_bodypart_names,
            )
        )
        pattern = r"(==|<=|>=|<|>)"
        all_events = []
        for relationship in animals_animals_relationships:
            for query in cross_animal_query_list:
                # assert that query must contain one of the operators
                # find the operator
                # note we need to strip the spaces
                comparison_operator = re.findall(pattern, query)[0].strip()
                _query = query.split(comparison_operator)[0].strip()

                _comparison = comparison_operator + "".join(
                    query.split(comparison_operator)[1:]
                )

                events = self.get_events_from_relationship(
                    relationship, _query, _comparison, smooth_window_size
                )
                all_events.extend(events)
        graph = EventGraph.init_from_list(all_events)

        ###

        if len(cross_animal_query_list) > 1:
            graphs = []
            for animal_name in self.animal_manager.get_animal_names():
                for receiver_animal_name in self.animal_manager.get_animal_names():
                    if animal_name != receiver_animal_name:
                        subgraph = EventGraph.fuse_subgraph_by_kvs(
                            graph,
                            {
                                "sender_animal_name": animal_name,
                                "receiver_animal_names": set([receiver_animal_name]),
                            },
                            number_of_overlap_for_fusion=len(cross_animal_query_list),
                        )
                        graphs.append(subgraph)

            graph = EventGraph.merge_subgraphs(graphs)

        ret_events = graph.to_list()
        ret_events = Event.filter_events_by_duration(ret_events, min_window, max_window)

        return ret_events

    @register_core_api
    def get_duration(self, events: List[Event]) -> int:
        """
        This function is for calculating the total duration of a list events.
        The return value is in seconds.
        """
        return sum([event.duration_in_seconds for event in events])

    @register_core_api
    def get_composite_events(
        self,
        events_A: List[Event],
        events_B: List[Event],
        composition_type: Literal[
            "sequential", "logical_and", "logical_or"
        ] = "logical_and",
        max_interval_between_sequential_events: int = 15,
        min_window: int = 15,
        max_window: int = 100000,
    ) -> List[Event]:
        """
        This function is for combining two sets of events.
        Parameters
        ----------
        events_A: The first events to combine. When composition type is 'sequential', this event happens first.
        events_B: The second events to combine. When composition type is 'sequential', this event happens second.
        max_interval_between_sequential_events: int, optional, default 15. Only used for 'sequential' composition type. The interval should be adjusted based on the nature of the two events.
        comopsition_type:
        'logical_and': two behaviors happen at the same time.
        'logical_or': one of the behaviors happen
        'sequential': the second behavior happens after the first behavior.

        Returns
        -------
        List[Event]
        """
        assert composition_type in [
            "sequential",
            "logical_and",
            "logical_or",
        ], "composition_type must be either 'sequential' or 'logical_or', or 'logical_and'"
        events_list = [events_A, events_B]
        if composition_type == "sequential":
            graph_list = [EventGraph.init_from_list(events) for events in events_list]
            graphs = []
            count = 0
            for animal_name in self.animal_manager.get_animal_names():
                for i in range(1, len(graph_list)):
                    count += 1
                    sub_graph = EventGraph.concat_graphs(
                        graph_list[i - 1],
                        graph_list[i],
                        {"sender_animal_name": animal_name},
                        max_interval_between_sequential_events,
                    )
                    graphs.append(sub_graph)

            graph = EventGraph.merge_subgraphs(graphs)

        elif composition_type == "logical_or":
            all_events = []
            for events in events_list:
                all_events.extend(events)
            ret = Event.filter_events_by_duration(all_events, min_window, max_window)
            return ret

        elif composition_type == "logical_and":
            all_events = []
            for idx, events in enumerate(events_list):
                all_events.extend(events)
            graph = EventGraph.init_from_list(all_events)
            graphs = []

            # we first fuse events from different task programs that involve animal-animal interactions
            for animal_name in self.animal_manager.get_animal_names():
                receiver_animal_names = [
                    animal_name
                    for animal_name in self.animal_manager.get_animal_names()
                ]
                for receiver_animal_name in receiver_animal_names:
                    if animal_name != receiver_animal_name:
                        receiver_animal_name = set([receiver_animal_name])
                        animal_animal_subgraph = EventGraph.fuse_subgraph_by_kvs(
                            graph,
                            {
                                "sender_animal_name": animal_name,
                                "receiver_animal_names": receiver_animal_name,
                            },
                            number_of_overlap_for_fusion=2,
                        )
                        graphs.append(animal_animal_subgraph)
            # we then fuse events from different task programs that involve animal-object interactions
            for object_name in self.object_manager.get_object_names():
                # if we strictly require the object needs to match, then EPM head dipping example won't work.
                # so we cannot require objects to match. This causes some ambiguity.
                animal_object_subgraph = EventGraph.fuse_subgraph_by_kvs(
                    graph,
                    {
                        "sender_animal_name": animal_name,
                    },
                    # "object_names": object_name},
                    number_of_overlap_for_fusion=2,
                )
                graphs.append(animal_object_subgraph)

            # fuse events from different task programs that involve animal states
            for animal_name in self.animal_manager.get_animal_names():
                animal_state_subgraph = EventGraph.handle_animal_state_fusion(
                    graph,
                    {
                        "sender_animal_name": animal_name,
                        "receiver_animal_names": set([]),
                    },
                )
                graphs.append(animal_state_subgraph)

            graph = EventGraph.merge_subgraphs(graphs)

        ret = graph.to_list()
        ret = Event.filter_events_by_duration(ret, min_window, max_window)
        return ret

    # @register_core_api
    def from_mask(
        self,
        mask_tensor: np.ndarray,
    ) -> List[Event]:
        """

        This function expects to take a binary mask to describe a condition for the behavior and returns a list of events.
        For animal-animal interaction, it expects shape of  (n_frames, n_individuals, n_individuals)
        For condition about animals' own states, it expects shape of (n_frames, n_individuals)
        Returns
        -------
        List[Event]
        """
        assert (
            len(mask_tensor.shape) == 2 or len(mask_tensor.shape) == 3
        ), "mask_tensor must be of shape (n_frames, n_individuals) or (n_frames, n_individuals, n_individuals)"
        # mask is of shape (n_frames, n_individuals)
        ret = []
        if len(mask_tensor.shape) == 2:
            for animal_id, animal_name in enumerate(
                self.animal_manager.get_animal_names()
            ):
                mask = mask_tensor[:, animal_id]
                events = Event.mask2events(
                    mask,
                    self.video_file_path,
                    animal_name,
                    set([]),
                    set([]),
                    smooth_window_size=3,
                )
                ret.extend(events)

        elif len(mask_tensor.shape) == 3:
            for animal_id, animal_name in enumerate(
                self.animal_manager.get_animal_names()
            ):
                for other_animal_id, other_animal_name in enumerate(
                    self.animal_manager.get_animal_names()
                ):
                    if animal_id == other_animal_id:
                        continue
                    mask = mask_tensor[:, animal_id, other_animal_id]

                    events = Event.mask2events(
                        mask,
                        self.video_file_path,
                        animal_name,
                        set([other_animal_name]),
                        set([]),
                        smooth_window_size=3,
                    )
                    ret.extend(events)
        return ret

    def get_serializeable_list_names(self) -> List[str]:
        return [
            "animals_object_events",
            "animals_animals_events",
            "animals_state_events",
        ]



from __future__ import annotations

import os
from typing import Any, Dict, List, Optional, Sequence, Set, Tuple, Union

import cv2
import numpy as np
from numpy import ndarray
from scipy.ndimage import uniform_filter1d

from amadeusgpt.analysis_objects.base import AnalysisObject


def moving_average(x: Sequence, window_size: int, pos: str = "centered"):
    """
    Compute the moving average of a time series.
    :param x: array_like, 1D input array
    :param window_size: int
        Must be odd positive, and less than or equal to the size of *x*.
    :param pos: str, optional (default="centered")
        Averaging window position.
        By default, the window is centered on the current data point,
        thus averaging over *window_size* // 2 past and future observations;
        no delay is introduced in the averaging process.
        Other options are "backward", where the average is taken
        from the past *window_size* observations; and "forward",
        where the average is taken from the future *window_size* observations.
    :return: ndarray
        Filtered time series with same length as input
    """
    # This function is not only very fast (unlike convolution),
    # but also numerically stable (unlike the one based on cumulative sum).
    # https://stackoverflow.com/questions/13728392/moving-average-or-running-mean/27681394#27681394
    x = np.asarray(x, dtype=float)
    x = np.squeeze(x)

    window_size = int(window_size)

    if window_size > x.size:
        raise ValueError("Window size must be less than or equal to the size of x.")

    if window_size < 1 or not window_size % 2:
        raise ValueError("Window size must be a positive odd integer.")

    middle = window_size // 2
    if pos == "centered":
        origin = 0
    elif pos == "backward":
        origin = middle
    elif pos == "forward":
        origin = -middle
    else:
        raise ValueError(f"Unrecognized window position '{pos}'.")

    return uniform_filter1d(x, window_size, mode="constant", origin=origin)


def smooth_boolean_mask(x: Sequence, window_size: int):
    # `window_size` should be at least twice as large as the
    # minimal number of consecutive frames to be smoothed out.
    if window_size % 2 == 0:
        window_size += 1
    return moving_average(x, window_size) > 0.5


def get_fps(video_path):
    # Load the video
    video = cv2.VideoCapture(video_path)
    # Get the FPS
    fps = video.get(cv2.CAP_PROP_FPS)
    video.release()
    return fps


def get_video_length(video_path):
    video = cv2.VideoCapture(video_path)
    fps = video.get(cv2.CAP_PROP_FPS)
    n_frames = video.get(cv2.CAP_PROP_FRAME_COUNT)
    return int(n_frames)


class BaseEvent(AnalysisObject):
    def __init__(
        self,
        start: int,
        end: int,
        video_file_path: str,
        data_length: int,
    ):
        # sometimes the frame length is not equal to the data length due to incomplete reading of the video file from pose
        self.video_file_path = video_file_path
        self.data_length = data_length
        self.start = start
        self.end = end
        assert os.path.exists(
            self.video_file_path
        ), f"the video file {video_file_path} does not exist"
        self.frame_length = get_video_length(self.video_file_path)

        duration_in_frames = self.end - self.start
        self.duration_in_frames = duration_in_frames
        duration_in_seconds = round(
            duration_in_frames / get_fps(self.video_file_path), 2
        )
        if duration_in_seconds < 0:
            self.duration_in_seconds = 0
        else:
            self.duration_in_seconds = round(
                duration_in_frames / get_fps(self.video_file_path), 2
            )
        self.duration = self.duration_in_frames

    def __len__(self) -> int:
        return self.end - self.start + 1

    def generate_mask(self) -> ndarray:
        temp = np.zeros(self.data_length, dtype=bool)
        temp[self.start : self.end + 1] = True
        return temp

    @classmethod
    def get_start_end_from_mask(cls, mask: ndarray) -> Tuple[int, int]:
        """
        Get the indices of first true and last true in the binary mask
        """
        start = int(np.argmax(mask))
        end = int(len(mask) - np.argmax(mask[::-1]) - 1)
        return start, end

    def __lt__(self, other):
        return self.duration_in_frames < other.duration_in_frames


class Event(BaseEvent):
    """
    Methods
    """

    def __init__(
        self,
        start: int,
        end: int,
        video_file_path: str,
        data_length: int,
        sender_animal_name: str,
        receiver_animal_names: Optional[Set[str]] = set(),
        object_names: Optional[Set[str]] = set(),
    ):
        # we allow only single sender animal for the event

        super().__init__(start, end, video_file_path, data_length)
        assert isinstance(
            sender_animal_name, str
        ), "sender_animal_name must be a string"
        self.sender_animal_name = sender_animal_name
        if receiver_animal_names is not None:
            assert isinstance(
                receiver_animal_names, set
            ), "receiver_animal_names must be a set"
        self.receiver_animal_names = receiver_animal_names
        if object_names is not None:
            assert isinstance(object_names, set), "object_names must be a set"
        self.object_names = object_names

    def __str__(self):
        return f"""
sender: {self.sender_animal_name}
receiver: {self.receiver_animal_names}
object: {self.object_names}
duration_in_seconds: {self.duration_in_seconds}
duration_in_frames: {self.duration_in_frames}
start: {self.start}
end: {self.end}
"""

    def __hash__(self):
        # Generate a hash based on a tuple of sorted (attribute, value) pairs
        # This accounts for all attributes dynamically
        items = tuple(sorted(self.__dict__.items()))
        return hash(items)

    def summary(self):
        ignore = ["mask", "video_file_path"]
        for attr_name in self.__dict__:
            if attr_name not in ignore:
                print(f"{attr_name}: {self.__dict__[attr_name]}")

    @classmethod
    def blockfy(cls, masks):
        blocks = []
        start = None
        for i, val in enumerate(masks):
            if val is not None and start is None:
                # Start of a new block
                start = i
            elif not val and start is not None:
                # End of a block
                if start != i - 1:
                    blocks.append((start, i - 1))
                start = None
        # If there is an open block at the end of the array, add it to the list
        if start is not None:
            blocks.append((start, len(masks) - 1))
        return blocks

    @classmethod
    # UNDER CONSTRUCTION
    def remove_overlapping_events(cls, events: List[Event]) -> List[Event]:
        # Assuming events are already sorted by start time
        filtered_events = []
        prev_event = None
        for event in events:
            if prev_event and event.end == prev_event.end:
                # Replace the previous event with the current one if they have the same end time
                filtered_events[-1] = event
            elif not prev_event or event.start >= prev_event.end:
                # Add the event if there is no previous event or it doesn't overlap
                filtered_events.append(event)

            prev_event = event
        return filtered_events

    @classmethod
    # UNDER CONSTRUCTION
    def event_negate(cls, events: List[Event]) -> List[Event]:
        """
        Get list of events that are negates of those events
        The core attributes must stay the same
        """
        mask = np.zeros(events[0].data_length, dtype=bool)
        sender_animal_name = events[0].sender_animal_name

        video_file_path = events[0].video_file_path
        for event in events:
            mask |= event.generate_mask()
        negate_mask = ~mask

        negate_events = Event.mask2events(
            negate_mask,
            video_file_path,
            sender_animal_name,
            set(),
            set(),
            smooth_window_size=1,
        )

        return negate_events

    @classmethod
    def check_max_in_sum(cls, events: List[Event]):
        """
        Check the maximum value in the sum of the masks of the events
        """
        masks = [event.generate_mask() for event in events]
        _sum = np.sum(masks, axis=0)
        return np.max(_sum)

    @classmethod
    def events2onemask(cls, events: List[Event]) -> ndarray:
        """
        return events to one mask that is a union of all the events
        """
        invovled_animals = set()
        for event in events:
            invovled_animals.add(event.sender_animal_name)
        assert (
            len(invovled_animals) <= 1
        ), f"This function must be called for events that have the same sender animal, you had {invovled_animals}"
        assert len(events) > 0, "events list is empty"

        mask = np.zeros(events[0].data_length, dtype=bool)
        for event in events:
            start, end = event.start, event.end
            mask[start : end + 1] = True
        return mask

    @classmethod
    def mask2events(
        cls,
        mask: ndarray,
        video_file_path: str,
        sender_animal_name: str,
        receiver_animal_names: Set[str],
        object_names: Set[str],
        smooth_window_size: int = 5,
    ) -> List[Event]:
        """
        Turn a binary mask to a list of Events
        The resulted event has the same sender animal name
        Each segment of the mask might have multiple receiver animals and object names
        Returns
        """
        if smooth_window_size is not None:
            mask = smooth_boolean_mask(mask, smooth_window_size)
        blocks = cls.blockfy(mask)
        events = []

        for block in blocks:
            start, end = block
            if end is not None:
                events.append(
                    Event(
                        start,
                        end,
                        video_file_path,
                        len(mask),
                        sender_animal_name,
                        receiver_animal_names=receiver_animal_names,
                        object_names=object_names,
                    )
                )

        return events

    @classmethod
    # UNDER CONSTRUCTION
    def filter_events_by_duration(
        cls,
        events: List[Event],
        min_duration: Union[int, float],
        max_duration: Union[int, float],
        unit: str = "frames",
    ) -> List[Event]:
        """
        Filter events by duration in frames
        """
        assert unit in ["frames", "seconds"], "must be either frames and seconds"

        if unit == "seconds":
            min_duration = int(min_duration * get_fps(events[0].video_file_path))
            max_duration = int(max_duration * get_fps(events[0].video_file_path))
            return [
                event
                for event in events
                if event.duration_in_seconds >= min_duration
                and event.duration_in_seconds <= max_duration
            ]
        else:
            min_duration = int(min_duration)
            max_duration = int(max_duration)

            return [
                event
                for event in events
                if event.duration_in_frames >= min_duration
                and event.duration_in_frames <= max_duration
            ]

    @classmethod
    # UNDER CONSTRUCTION
    def concat_two_events(
        cls,
        early_event: Event,
        late_event: Event,
    ) -> Union[None, Event]:
        """
        Concatenate two events into one, fill the gap between two events if there are
        """
        assert early_event.sender_animal_name == late_event.sender_animal_name

        object_names = early_event.object_names
        receiver_animal_names = early_event.receiver_animal_names

        new_event = Event(
            early_event.start,
            late_event.end,
            early_event.video_file_path,
            early_event.data_length,
            early_event.sender_animal_name,
            receiver_animal_names=receiver_animal_names,
            object_names=object_names,
        )
        return new_event


# Trying to implement temporal graph
class Node:
    def __init__(self, start: int, children: List[Event]):
        self.children = children
        self.start: int = start
        self.next: Node | None = None
        self.prev: Node | None = None

    def get_children_by_key(self, key: str):
        """ """
        assert key in ["sender_animal_name", "receiver_animal_name", "object_name"]
        return [child for child in self.children if child.key == key]

    @classmethod
    def copy(cls, node):
        ret = cls(node.start, node.children)
        ret.prev = None
        ret.next = None
        return ret


class EventGraph:
    """
    Implementing a temporal graph
    Every node keeps a collection of events that have the same start time

    """

    def __init__(self):
        self.head: Node = None
        self.n_nodes = 0

    def to_list(self) -> List[Event]:
        ret = []
        cur_node = self.head
        while cur_node is not None:
            ret.extend(cur_node.children)
            cur_node = cur_node.next
        ret = sorted(ret, key=lambda x: x.start)
        ret = [e for e in ret if e.duration_in_frames > 1]

        return ret

    @property
    def animal_names(self):
        return list(set([event.sender_animal_name for event in self.to_list()]))

    @classmethod
    def check_list_sorted(cls, events: List[Event]) -> bool:
        """
        Check if the list of events is sorted
        """
        temp = [event.start for event in events]
        return temp == sorted(temp)

    @classmethod
    def init_from_list(cls, events: List[Event]) -> "EventGraph":
        graph = cls()
        start_time_dict = {}

        for event in events:
            start = event.start
            if start in start_time_dict:
                start_time_dict[start].append(event)
            else:
                start_time_dict[start] = [event]

        for start_time in start_time_dict:

            node = Node(start_time, start_time_dict[start_time])
            graph.insert_node(node)

        return graph

    @classmethod
    def init_from_mask(
        cls,
        mask: ndarray,
        video_file_path: str,
        sender_animal_name: str,
        receiver_animal_names: Set[str],
        object_names: Set[str],
    ) -> "EventGraph":

        events = Event.mask2events(
            mask,
            video_file_path,
            sender_animal_name,
            receiver_animal_names=receiver_animal_names,
            object_names=object_names,
        )

        graph = cls.init_from_list(events)

        return graph

    def insert_node(self, new_node: Node):
        """
        Insert a new node into the graph
        We greedily insert the node after the first node that has a start time less than the new node
        We leave the follow-up operations in other functions
        """
        # we might reuse those nodes so we need to clean them

        new_node = Node.copy(new_node)
        # find the right place to insert the node and update the linked list
        # empty graph

        if self.head is None:
            self.head = new_node
            self.n_nodes += 1
            return

        # only one node in the graph

        if new_node.start < self.head.start:
            new_node.next = self.head
            self.head.prev = new_node
            self.head = new_node
            return
        if self.head.next is None:
            if new_node.start == self.head.start:
                self.head.children.extend(new_node.children)
            else:
                self.head.next = new_node
                new_node.prev = self.head
            self.n_nodes += 1
            return

        cur_node = self.head

        while cur_node.next is not None:
            if new_node.start > cur_node.start and new_node.start < cur_node.next.start:
                new_node.next = cur_node.next
                new_node.next.prev = new_node
                cur_node.next = new_node
                new_node.prev = cur_node
                self.n_nodes += 1
                return
            elif cur_node.start == new_node.start:
                # let's check not to insert a basically equivalent node
                cur_node.children.extend(new_node.children)
                self.n_nodes += 1
                return
            else:
                cur_node = cur_node.next
        # the new node is the last node

        cur_node.next = new_node
        new_node.prev = cur_node

        self.n_nodes += 1

    @classmethod
    def animal_group_subgraph(cls, graph: "EventGraph", n_individuals: int):
        """
        WIP
        Retrieve the subgraph where multiple animals satisfy the same condition at the same time.
        """
        events = graph.to_list()
        animals = set([event.sender_animal_name for event in events])
        masks = []
        for idx, animal_name in enumerate(animals):
            events_animal = [
                event for event in events if event.sender_animal_name == animal_name
            ]
            # find the overlapping between events_animal_a and events_animal_b
            mask = Event.events2onemask(events_animal)
            masks.append(mask)
        masks = np.concatenate(masks, axis=0)
        _sum = np.sum(masks, axis=0)
        ret = np.zeros_like(mask, dtype=bool)
        ret[_sum == n_individuals] = True
        # get where _sum
        return cls.init_from_mask(
            ret, events[0].video_file_path, events[0].sender_animal_name, set(), set()
        )

    @classmethod
    def refine_graph(cls, graph):
        # 1) when animal state is present, the receiver name is an empty set. Merge those events with their closest neighbors
        # 2) when multiple events share the sender name, the same start and end time but differ in receiver name and object name, we can create a merged event
        pass

    @classmethod
    def handle_animal_state_fusion(cls, graph, merge_kvs: Dict[str, Any]):
        # only matching animal state events that have no receiver name or object name
        animal_state_events = graph.traverse_by_kvs(merge_kvs)
        sender_animal_name = merge_kvs["sender_animal_name"]
        new_graph = cls()
        events = graph.to_list()
        for event in events:
            # identify non animal state events
            if (
                event.sender_animal_name == sender_animal_name
                and len(event.receiver_animal_names) != 0
            ):
                for animal_state_event in animal_state_events:
                    start1, end1 = event.start, event.end
                    start2, end2 = animal_state_event.start, animal_state_event.end
                    if max(start1, start2) <= min(end1, end2):
                        mask1 = event.generate_mask()
                        mask2 = animal_state_event.generate_mask()
                        mask = mask1 & mask2
                        # animal state event keeps the original receiver name and object name of the non-state event
                        true_indices = np.where(mask)[0]

                        new_event = Event(
                            true_indices[0],
                            true_indices[-1],
                            event.video_file_path,
                            len(mask),
                            sender_animal_name=event.sender_animal_name,
                            receiver_animal_names=event.receiver_animal_names,
                            object_names=event.object_names,
                        )
                        new_graph.insert_node(Node(new_event.start, [new_event]))
        return new_graph

    @classmethod
    def fuse_subgraph_by_kvs(
        cls,
        graph: "EventGraph",
        merge_kvs: Dict[str, Any],
        number_of_overlap_for_fusion: int = 0,
        allow_more_than_2_overlap: bool = False,
    ) -> "EventGraph":
        """
        number_of_overlap_for_fusion is a parameter for logical and.
        For example, if there are two conditions to be met in the masks we look for locations that have overlap as 2
        """
        # retrieve all events that satisfy the conditions (k=v)
        events = graph.traverse_by_kvs(merge_kvs)
        if not allow_more_than_2_overlap:
            assert (
                Event.check_max_in_sum(events) <= number_of_overlap_for_fusion
            ), f"Detected overlap {Event.check_max_in_sum(events)}. But we only allow {number_of_overlap_for_fusion} overlap for fusion."

        new_graph = cls()
        if len(events) == 0:
            return new_graph

        masks = [event.generate_mask() for event in events]
        _sum = np.sum(masks, axis=0)
        mask = np.zeros_like(_sum, dtype=bool)
        # in case there are many events overlap,
        # if the events come from one single condition, there is no overlap
        # so the overlap must come from different conditions
        mask[(_sum >= number_of_overlap_for_fusion)] = True

        # we must be extra careful when we do this
        # assuming traversing by kv enforces the exact match of receiver name and object name
        # it's safe to assign receiver name and object name from the first event
        events = Event.mask2events(
            mask,
            events[0].video_file_path,
            events[0].sender_animal_name,
            receiver_animal_names=events[0].receiver_animal_names,
            object_names=events[0].object_names,
        )

        for event in events:
            new_graph.insert_node(Node(event.start, [event]))

        return new_graph

    @classmethod
    def merge_subgraphs(cls, graph_list: List["EventGraph"]) -> "EventGraph":
        """
        Merge graphs into one graph. There is no fusion. Nodes are just added based on the start time
        """
        graph = cls()
        for g in graph_list:
            head = g.head
            while head is not None:
                graph.insert_node(head)
                head = head.next
        return graph

    def traverse_by_kvs(self, kvs: Dict[str, Any]) -> List[Event]:
        """
        Traverse the graph and get a list of events that have the same key
        """
        ret = []
        cur_node = self.head
        while cur_node is not None:
            for event in cur_node.children:
                conditions = []
                for k, v in kvs.items():
                    # we do more strict exact match here from now
                    conditions.append(getattr(event, k) == v)
                if all(conditions):
                    ret.append(event)

            cur_node = cur_node.next
        return ret

    @classmethod
    def concat_graphs(
        cls,
        graph1: "EventGraph",
        graph2: "EventGraph",
        merge_kvs: Dict[str, Any],
        max_interval_between_sequential_events,
    ) -> "EventGraph":
        """
        Concatenate two graphs in a sequential manner
        For every node in the concated graph, it must be a node from grpah1 followed by a node from graph2
        """
        graph = cls()

        events1 = graph1.traverse_by_kvs(merge_kvs)
        events2 = graph2.traverse_by_kvs(merge_kvs)

        secure_events1 = []
        secure_events2 = []

        for i, event2 in enumerate(events2):
            for j, event1 in enumerate(events1):

                # skip if event1 is not the closest event to event2
                if j < len(events1) - 1 and events1[j + 1].end <= event2.start:
                    continue
                else:
                    # this is strict continuous
                    # if event1.end >= event2.start and event1.end <= event2.end:
                    # this is not strict continuous
                    if (
                        event1.duration_in_frames > 0
                        and event1.end <= event2.start
                        and event2.start - event1.end
                        <= max_interval_between_sequential_events
                    ):
                        secure_events1.append(event1)
                        secure_events2.append(event2)
                        new_event = Event.concat_two_events(event1, event2)
                        graph.insert_node(Node(new_event.start, [new_event]))
                        break

        return graph

    def display_graph(self):
        # print ('asked to display')
        cur_node = self.head
        while cur_node is not None:
            for event in cur_node.children:
                print(event.start, event.end)

            cur_node = cur_node.next



import ast
import inspect
import sys
import time
import traceback
from collections import defaultdict

import numpy as np

from amadeusgpt.analysis_objects.event import Event
from amadeusgpt.logger import AmadeusLogger


def filter_kwargs_for_function(func, kwargs):
    sig = inspect.signature(func)
    return {k: v for k, v in kwargs.items() if k in sig.parameters}


def timer_decorator(func):
    def wrapper(*args, **kwargs):
        start_time = time.time()  # before calling the function
        result = func(*args, **kwargs)  # call the function
        end_time = time.time()  # after calling the function
        AmadeusLogger.debug(
            f"The function {func.__name__} took {end_time - start_time} seconds to execute."
        )
        print(
            f"The function {func.__name__} took {end_time - start_time} seconds to execute."
        )
        return result

    return wrapper


def parse_error_message_from_python():
    exc_type, exc_value, exc_traceback = sys.exc_info()
    traceback_str = "".join(
        traceback.format_exception(exc_type, exc_value, exc_traceback)
    )
    return traceback_str


def validate_openai_api_key(key):
    import openai

    openai.api_key = key
    try:
        openai.models.list()
        return True
    except openai.AuthenticationError:
        return False


def flatten_tuple(t):
    """
    Used to handle function returns
    """
    flattened = []
    for item in t:
        if isinstance(item, tuple):
            flattened.extend(flatten_tuple(item))
        else:
            flattened.append(item)
    return tuple(flattened)


# the function func2json takes a function object as inputs
# and returns a json object that contains the function name,
# input type, output type, function body, and function description

import inspect
import textwrap


def func2json(func):
    if isinstance(func, str):
        func_str = textwrap.dedent(func)

        # Parse the function string to an AST
        parsed = ast.parse(func_str)
        func_def = parsed.body[0]
        # Use the AST to extract the function's name
        func_name = func_def.name
        # Extract the docstring directly from the AST
        docstring = ast.get_docstring(func_def)
        # Remove the docstring node if present
        if (
            func_def.body
            and isinstance(func_def.body[0], ast.Expr)
            and isinstance(func_def.body[0].value, (ast.Str, ast.Constant))
        ):
            func_def.body.pop(0)

        # Remove decorators from the function definition
        func_def.decorator_list = []

        # Convert the modified AST back to source code
        if hasattr(ast, "unparse"):
            source_without_docstring_or_decorators = ast.unparse(func_def)
        else:
            # Placeholder for actual conversion in older Python versions
            source_without_docstring_or_decorators = None  # Consider using `astor` here

        # Attempt to evaluate return annotation, if any
        return_annotation = "No return annotation"
        if func_def.returns:
            return_annotation = ast.unparse(func_def.returns)

        json_obj = {
            "name": func_name,
            "inputs": "",
            "source_code": source_without_docstring_or_decorators,
            "docstring": docstring,
            "return": return_annotation,
        }
        return json_obj
    else:

        # Capture the function's signature for input arguments
        sig = inspect.signature(func)
        inputs = {name: str(param.annotation) for name, param in sig.parameters.items()}

        # Extract the docstring
        docstring = inspect.getdoc(func)
        if docstring:
            docstring = textwrap.dedent(docstring)

        # Capture the function's full source and parse it to an AST
        full_source = inspect.getsource(func)
        parsed = ast.parse(textwrap.dedent(full_source))
        func_def = parsed.body[0]

        # Remove the docstring node if present
        if (
            func_def.body
            and isinstance(func_def.body[0], ast.Expr)
            and isinstance(func_def.body[0].value, (ast.Str, ast.Constant))
        ):
            func_def.body.pop(0)

        # Remove decorators from the function definition
        func_def.decorator_list = []

        # Convert the modified AST back to source code
        if hasattr(ast, "unparse"):
            source_without_docstring_or_decorators = ast.unparse(func_def)
        else:
            # Placeholder for actual conversion in older Python versions
            source_without_docstring_or_decorators = None  # Consider using `astor` here

        json_obj = {
            "name": func.__name__,
            "inputs": inputs,
            "source_code": textwrap.dedent(source_without_docstring_or_decorators),
            "docstring": docstring,
            "return": str(sig.return_annotation),
        }
        return json_obj


class QA_Message:
    def __init__(self, query: str, video_file_paths: list[str]):
        # user question
        self.query = query
        self.video_file_paths = video_file_paths
        # llm generated code
        self.code = None
        self.chain_of_thought = None
        ### following fields change per video
        # a reference to the sandbox
        self.error_message = defaultdict(list)
        self.plots = defaultdict(list)
        self.out_videos = defaultdict(list)
        self.pose_video = defaultdict(list)
        self.function_rets = defaultdict(list)
        self.meta_info = {}

    def get_masks(self) -> dict[str, np.ndarray]:
        ret = {}
        function_rets = self.function_rets
        # if function_ret is a list of events
        for video_path, rets in function_rets.items():
            if isinstance(rets, list) and len(rets) > 0 and isinstance(rets[0], Event):
                events = rets
                masks = []
                for event in events:
                    masks.append(event.generate_mask())
                ret[video_path] = np.array(masks)
            else:
                ret[video_path] = None

        return ret

    # STILL UNDER CONSTRUCTION. NEED TO HANDLE SERIALIZATION OF PLOTS
    def serialize_qa_message(self):
        return {
            "query": self.query,
            "video_file_paths": self.video_file_paths,
            "code": self.code,
            "chain_of_thought": self.chain_of_thought,
            "error_message": self.error_message,
            "plots": None,
            "out_videos": self.out_videos,
            "pose_video": self.pose_video,
            "function_rets": self.function_rets,
            "meta_info": self.meta_info,
        }


def create_qa_message(query: str, video_file_paths: list[str]) -> QA_Message:

    return QA_Message(query, video_file_paths)


from IPython.display import Markdown, Video, display


def parse_result(amadeus, qa_message, use_ipython=True, skip_code_execution=False):
    if use_ipython:
        display(Markdown(qa_message.chain_of_thought))
    else:
        print(qa_message.chain_of_thought)
    sandbox = amadeus.sandbox
    if not skip_code_execution:
        qa_message = sandbox.code_execution(qa_message)
    qa_message = sandbox.render_qa_message(qa_message)
    if len(qa_message.out_videos) > 0:
        print(f"videos generated to {qa_message.out_videos}")
        print(
            "Open it with media player if it does not properly display in the notebook"
        )
        if use_ipython:
            if len(qa_message.out_videos) > 0:
                for video_path, event_videos in qa_message.out_videos.items():
                    for event_video in event_videos:
                        display(Video(event_video, embed=True))

    if use_ipython:
        if len(qa_message.function_rets) > 0:
            for video_file_path in qa_message.function_rets:
                display(Markdown(str(qa_message.function_rets[video_file_path])))

    return qa_message



"""
This class is for deep learning models
"""

import os
import pickle
import platform
from typing import Any, Dict, List, Union

import cv2
import msgpack
import numpy as np

from .base import AnalysisObject
from .object import Object


def _superanimal_inference(
    self, video_file_path, superanimal_name, scale_list, video_adapt
):
    import deeplabcut

    progress_obj = st.progress(0)
    deeplabcut.video_inference_superanimal(
        [video_file_path],
        superanimal_name,
        scale_list=scale_list,
        progress_obj=progress_obj,
        video_adapt=True,
        pseudo_threshold=0.5,
    )


def superanimal_video_inference(
    self,
    superanimal_name="superanimal_topviewmouse",
    scale_list=[],
    video_adapt=False,
):
    """
    Examples
    --------
     # extract pose from the video file with superanimal name superanimal_topviewmouse
     def task_program():
         superanimal_name = "superanimal_topviewmouse"
         keypoint_file_path = AnimalBehaviorAnalysis.superanimal_video_inference(superanimal_name)
         return keypoint_file_path
    """

    import glob

    if "streamlit_cloud" in os.environ:
        raise NotImplementedError(
            "Due to resource limitation, we do not support superanimal inference in the app"
        )

    video_file_path = type(self).get_video_file_path()

    self._superanimal_inference(
        video_file_path, superanimal_name, scale_list, video_adapt
    )

    vname = Path(video_file_path).stem
    resultfolder = Path(video_file_path).parent
    # resultfile should be a h5
    # right now let's consider there is only one file
    # in the future we need to consider multiple files
    print("resultfolder", resultfolder)
    resultfile = glob.glob(os.path.join(resultfolder, vname + "DLC*.h5"))[0]
    print("resultfile", resultfile)
    if os.path.exists(resultfile):
        Database.add(type(self).__name__, "keypoint_file_path", resultfile)

    else:
        raise ValueError(f"{resultfile} not exists")

    pose_video_file = resultfile.replace(".h5", "_labeled.mp4")
    Database.add("AnimalBehaviorAnalysis", "pose_video_file", pose_video_file)

    return pose_video_file


from segment_anything import (SamAutomaticMaskGenerator, SamPredictor,
                              sam_model_registry)


class Model(AnalysisObject):
    def __init__(self, config):
        self.config = config


class DeepLabCut(Model):
    pass


class Segmentation(Model):
    """
    Base class for segmentation.
    Should support saving the mask to disk and loading it automatically
    This is because model like SAM can take a long time
    """

    def __init__(self, config: Dict[str, any]):
        """
        filename specifies the path to the potential serialized segmentation file
        We make sure that the segmentation files have same formats
        """
        self.filename = config["seg_filename"]
        self.pickledata = None
        self.load()

    def get_name(self) -> str:
        return self.filename

    def load_msgpack(self):
        object_list = {
            0: "barrel",
            1: "cotton",
            2: "food",
            3: "igloo",
            4: "tower",
            5: "tread",
            6: "tunnel",
            7: "water",
        }

        with open(self.filename, "rb") as f:
            unpacker = msgpack.Unpacker(f, raw=False)
            print("loading seg from maushaus file")
            for frame_id, data_at_frame in enumerate(unpacker):
                mask_dict = {}
                for object in data_at_frame:
                    assert frame_id == object["frame_id"]
                    object_name = object_list[object["category_id"]]
                    bbox = object["bbox"]
                    # because maushaus does not have area, I calculate it from bbox
                    x, y, w, h = bbox
                    image_size = object["segmentation"]["size"]
                    # try not to evaluate the string
                    mask_dict[object_name] = {
                        "segmentation": object["segmentation"],
                        "area": w * h,
                        "bbox": bbox,
                    }
                break
            # now let's just use the first frame
            self.pickledata = mask_dict

    def load_pickle(self):
        if os.path.exists(self.filename):
            with open(self.filename, "rb") as f:
                self.pickledata = pickle.load(f)

    def load(self):
        if self.filename is not None:
            if self.filename.endswith("msgpack"):
                self.load_msgpack()
            elif self.filename.endswith("pickle"):
                self.load_pickle()
            else:
                raise ValueError(f"{self.filename} not supported")

    def save_to_pickle(self, data, filename):
        with open(filename, "wb") as f:
            pickle.dump(data, f)


class SAM(Segmentation):
    """
    Class that captures the state of objects, supported by Seg everything
    """

    # def __init__(self, ckpt_path, model_type, filename=None):
    def __init__(self, sam_info: Dict[str, Any]):
        super().__init__(sam_info)
        self.sam_info = sam_info
        self.ckpt_path: Union[str, None] = self.sam_info.get("ckpt_path")
        self.model_type: Union[str, None] = self.sam_info.get("model_type")
        self.scene_frame_number = self.sam_info.get("scene_frame_number")

        sam = sam_model_registry[self.model_type](checkpoint=self.ckpt_path)
        device = "cpu" if platform.system() == "Darwin" else "cuda"
        sam.to(device)
        self.mask_generator = SamAutomaticMaskGenerator(sam)

    def generate_mask(self, image):
        masks = self.mask_generator.generate(image)
        return masks

    def generate_mask_at_frame(self, video_file_path, frame_id):
        cap = cv2.VideoCapture(video_file_path)
        cap.set(cv2.CAP_PROP_POS_FRAMES, frame_id)
        ret, frame = cap.read()
        masks = self.generate_mask(frame)
        cap.release()
        # cv2.destroyAllWindows()
        return masks

    def get_objects(self, video_file_path: str, sam_info: Dict[str, Any], frame_id=0):
        # assuming objects are still
        if self.pickledata is None:
            masks = self.generate_mask_at_frame(video_file_path, frame_id)
            objects = {}
            for object_name, mask in enumerate(masks):
                obj = Object(str(object_name), masks=mask)
                objects[str(object_name)] = obj
            return objects
        else:
            return self.pickledata


class MausHausSeg(Segmentation):
    def __init__(self, filename=None):
        super().__init__(filename=filename)

    def get_objects(self):
        ret = {}
        if self.pickledata is not None:
            print("building maushaus objects from rle string")
            for object_name, masks in self.pickledata.items():
                ret[object_name] = Object(object_name, masks=masks)
            return ret
        else:
            raise ValueError("We only support loading from MausHaus for now")



"""
Define a series of transform that embedding algorithms might find usesful
Not sure yet how much this can be done by GPT-4 or done by us
"""

import numpy as np


def align_poses(self, keypoints: np.ndarray) -> np.ndarray:
    """
    Aligns 2D keypoints by translating them to the mean position and scaling them based on the bounding box size.

    Args:
        keypoints (numpy.ndarray): 2D array of keypoints with shape (n_frames, n_kpts, 2).

    Returns:
        numpy.ndarray: Aligned keypoints with the same shape as the input.
    """
    n_frames, n_individuals, n_kpts, _ = keypoints.shape
    aligned_keypoints = np.copy(keypoints)
    # Compute mean position across all frames
    mean_position = np.mean(aligned_keypoints, axis=(0, 2))
    # Translate keypoints to mean position
    aligned_keypoints -= mean_position
    # Compute maximum absolute coordinate value across all frames
    max_abs_coord = np.max(np.abs(aligned_keypoints))
    # Scale keypoints based on bounding box size
    aligned_keypoints /= max_abs_coord

    return aligned_keypoints



import matplotlib.path as mpath
import numpy as np

from .base import AnalysisObject


class Object(AnalysisObject):
    def __init__(self, name: str):
        """
        TODO: instead of using a point, use the true segmentation as reference point
        name: str for referencing the object
        segmentation : the mask
        area : the area of the mask in pixels
        bbox : the boundary box of the mask in XYWH format
        predicted_iou : the model's own prediction for the quality of the mask
        point_coords : the sampled input point that generated this mask
        stability_score : an additional measure of mask quality
        crop_box : the crop of the image used to generate this mask in XYWH format

        Attributes
        ----------
        center: x,y the center of the object
        """
        self.name = name
        self.center = None
        self.x_min = None
        self.x_max = None
        self.y_min = None
        self.y_max = None
        self.Path = None
        self.points = None

    def get_name(self):
        return self.name

    def get_data(self):
        """
        The class multiple fields
        """
        return self

    def get_center(self):
        return self.center

    def get_xmin(self):
        return self.x_min

    def get_xmax(self):
        return self.x_max

    def get_ymin(self):
        return self.y_min

    def get_ymax(self):
        return self.y_max

    def is_valid(self):
        # subclass can overwrite it
        return True

    def __getitem__(self, key):
        return getattr(self, key)

    def summary(self):
        for attr_name in self.__dict__:
            print(f"{attr_name} has {self.__dict__[attr_name]}")

    def distance(self, other_object):
        # we use the center of two objects for calculating distance
        return np.linalg.norm(self.center - other_object.center)

    def points2Path(self, points):
        path = None
        if len(points) > 2:
            # hull = ConvexHull(points)
            # points are in counter-clockwise order
            # vertices = hull.vertices.astype(np.int64)
            # points are mesh points. We need to convert them to convex hull
            path_data = []
            path_data.append((mpath.Path.MOVETO, points[0]))
            for point in points[1:]:
                path_data.append((mpath.Path.LINETO, point))
            path_data.append((mpath.Path.CLOSEPOLY, points[0]))
            codes, verts = zip(*path_data)
            path = mpath.Path(verts, codes)
        return path

    def get_path(self):
        """
        The representation of points in the object
        """
        return self.Path

    def overlap(self, other_object):
        # if there is a polygon path corresponding this object, use the contain_point
        # otherwise, use the bounding box representation
        if self.get_path() is not None:
            return self.path.contains_point(other_object.center)
        else:
            other_object_center = other_object.center
            return (
                other_object_center <= self.x_max
                and other_object_center <= self.y_max
                and other_object_center >= self.x_min
                and other_object_center >= self.y_min
            )

    def to_left(self, other_object):
        # whether the other object is to the left of this object
        return other_object.center[0] <= self.x_min

    def to_right(self, other_object):
        # whether the other object is to the right of this object
        return other_object.center[0] >= self.x_max

    def to_above(self, other_object):
        # whether the other object is to the above of this object
        return other_object.center[1] <= self.y_min

    def to_below(self, other_object):
        # whether the other object is to the below of this object
        return other_object.center[1] >= self.y_max


class SegObject(Object):
    """
    segmentation : the mask
    area : the area of the mask in pixels
    bbox : the boundary box of the mask in XYWH format
    predicted_iou : the model's own prediction for the quality of the mask
    point_coords : the sampled input point that generated this mask
    stability_score : an additional measure of mask quality
    crop_box : the crop of the image used to generate this mask in XYWH format
    """

    def __init__(self, name: str, masks: dict):
        super().__init__(name)
        self.masks = masks
        self.bbox = self.masks.get("bbox")
        self.area = self.masks["area"]
        # _seg could be either binary mask or rle string
        _seg: dict = self.masks.get("segmentation")
        # this is rle format
        from pycocotools import mask as mask_decoder

        if "counts" in _seg:
            self.segmentation = mask_decoder.decode(_seg)
        else:
            self.segmentation = masks.get("segmentation")
        point_coords = np.where(self.segmentation)
        point_coords = zip(point_coords[0], point_coords[1])
        # need to revert x and y for matplotlib plotting
        points = [[p[1], p[0]] for p in point_coords]
        points = np.array(points)
        self.points = points
        self.Path = self.points2Path(points)
        x, y, w, h = self.bbox
        self.x_min, self.y_min, self.x_max, self.y_max = x, y, x + w, y + h
        self.center = np.array([x + w / 2, y + h / 2])

    def plot(self, ax):
        x, y = zip(*self.points)
        # Plot the polygon
        ax.plot(x, y, "b-")  # 'b-' means blue line


class MatplotlibObject(Object):
    def __init__(self, name: str, object_path):
        super().__init__(name)
        self.object_path = object_path
        self.Path = object_path
        vertices = self.Path.vertices
        self.x_min = np.nanmin(vertices[:, 0])
        self.x_max = np.nanmax(vertices[:, 0])
        self.y_min = np.nanmin(vertices[:, 1])
        self.y_max = np.nanmax(vertices[:, 1])
        # Calculate the area of the convex hull using the Shoelace formula
        x = vertices[:, 0]
        y = vertices[:, 1]
        self.points = vertices
        self.area = 0.5 * np.abs(np.dot(x, np.roll(y, 1)) - np.dot(y, np.roll(x, 1)))
        self.center = np.array([np.mean(vertices[:, 0]), np.mean(vertices[:, 1])])


class ROIObject(Object):
    def __init__(self, name: str, canvas_path):
        super().__init__(name)

        self.canvas_path = canvas_path
        points = canvas_path
        if isinstance(canvas_path, mpath.Path):
            self.Path = canvas_path
        else:
            self.Path = self.points2Path(points)
        vertices = self.Path.vertices
        self.x_min = np.nanmin(vertices[:, 0])
        self.x_max = np.nanmax(vertices[:, 0])
        self.y_min = np.nanmin(vertices[:, 1])
        self.y_max = np.nanmax(vertices[:, 1])
        # Calculate the area of the convex hull using the Shoelace formula
        x = vertices[:, 0]
        y = vertices[:, 1]
        self.points = vertices
        self.area = 0.5 * np.abs(np.dot(x, np.roll(y, 1)) - np.dot(y, np.roll(x, 1)))
        self.center = np.array([np.mean(vertices[:, 0]), np.mean(vertices[:, 1])])


class GridObject(Object):
    def __init__(self, name: str, region):
        super().__init__(name)
        self.region = region
        self.x_min = region["x"]
        self.y_min = region["y"]
        self.x_max = region["x"] + region["w"]
        self.y_max = region["y"] + region["h"]
        self.center = np.array(
            [self.x_min + region["w"] / 2, self.y_min + region["h"] / 2]
        )

        self.Path = self.points2Path(
            [
                [self.x_min, self.y_min],
                [self.x_max, self.y_min],
                [self.x_max, self.y_max],
                [self.x_min, self.y_max],
            ]
        )
        self.points = np.array(
            [
                [self.x_min, self.y_min],
                [self.x_max, self.y_min],
                [self.x_max, self.y_max],
                [self.x_min, self.y_max],
            ]
        )
        self.area = region["w"] * region["h"]



import base64
import gc
import glob
import io
import json
import os
import tempfile
from io import BytesIO
from pathlib import Path

import cv2
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import requests
import streamlit as st
from PIL import Image
from streamlit_drawable_canvas import st_canvas

import amadeusgpt
from amadeusgpt import AMADEUS
from amadeusgpt.analysis_objects.object import Object, ROIObject
from amadeusgpt.behavior_analysis.analysis_factory import create_analysis
from amadeusgpt.behavior_analysis.identifier import Identifier
from amadeusgpt.config import Config
from amadeusgpt.utils import QA_Message

LOG_DIR = os.path.join(os.path.expanduser("~"), "Amadeus_logs")
VIDEO_EXTS = "mp4", "avi", "mov"
current_script_directory = os.path.dirname(os.path.abspath(__file__))
user_profile_path = os.path.join(
    current_script_directory, "static", "images", "cat.png"
)
bot_profile_path = os.path.join(
    current_script_directory, "static", "images", "chatbot.png"
)


def load_profile_image(image_path):
    if image_path.startswith("http"):
        response = requests.get(image_path)
        img = Image.open(BytesIO(response.content))
    else:
        img = Image.open(image_path)
    return img


def load_css():
    current_script_directory = os.path.dirname(os.path.abspath(__file__))
    css_path = os.path.join(current_script_directory, "static/styles/style.css")
    if os.path.exists(css_path):
        st.markdown(f"<style>{open(css_path).read()}</style>", unsafe_allow_html=True)
    else:
        st.error(f"File not found: {css_path}")


USER_PROFILE = load_profile_image(user_profile_path)
BOT_PROFILE = load_profile_image(bot_profile_path)


class BaseMessage:
    def __init__(self, json_entry=None):

        self.role = ""
        self.query = ""
        if json_entry:
            json_entry = self.parse_from_json_entry(json_entry)
            self.role = json_entry["role"]
            self.query = json_entry["query"]

    def parse_from_json_entry(self, json_entry):
        return json_entry

    def format_caption(self, caption):
        temp = caption.split("\n")
        temp = [f"<li>{content}</li>" for content in temp]
        ret = "<ul>\n" + "".join(temp).rstrip() + "\n</ul>"
        return ret

    def render(self):
        raise NotImplementedError("Must implement this")


class HumanMessage(BaseMessage):
    def __init__(self, query=None, json_entry=None):
        if json_entry:
            super().__init__(json_entry=json_entry)
        self.role = "human"
        if query:
            self.query = query

    def render(self):
        st.markdown(
            f'<div class="panel">{self.query}</div>',
            unsafe_allow_html=True,
        )


class AIMessage(BaseMessage):
    def __init__(self, qa_message: QA_Message | None, json_entry=None):
        self.rendered = False
        if json_entry:
            super().__init__(json_entry=json_entry)

        self.role = "ai"

        self.query = qa_message.query
        self.chain_of_thought = qa_message.chain_of_thought
        self.error_message = qa_message.error_message
        self.function_rets = qa_message.function_rets
        self.plots = qa_message.plots
        self.out_videos = qa_message.out_videos
        self.code = qa_message.code
        self.pose_video = qa_message.pose_video

    def render(self, debug=False):
        """
        We use the getter for better encapsulation
        overall structure of what to be rendered

              | chain of thoughts
        error | str_answer | nd_array
              | helper_code
              | code
              | figure
              |      | figure_explanation
              | figure
              |      | figure_explanation
              | overall_explanation

        --------
        """

        render_keys = [
            "query",
            "chain_of_thought",
            "error_message",
            "function_rets",
            "plots",
            "out_videos",
        ]
        # for render_key, render_value in self.data.items():

        for render_key in render_keys:
            if getattr(self, render_key, None) is None:
                continue
            render_value = getattr(self, render_key)
            if render_value is None:
                # skip empty field
                continue
            if render_key == "function_rets":
                if isinstance(render_key, (list, tuple)):
                    for ret in render_value:
                        for key, value in ret.items():
                            st.code(f"Result\n {value}\n ", language="python")
                else:
                    st.code(f"Result\n {render_value}\n ", language="python")
            elif render_key == "error_message":
                error_message = render_value
                for video_file_path, error in error_message.items():
                    if error is not None:
                        st.markdown(f"The error says: {error}\n ")
            elif render_key == "chain_of_thought":
                # there should be a better matching than this
                text = render_value
                lines = text.split("\n")
                inside_code_block = False
                code_block = []
                for line in lines:
                    if line.strip().startswith("```python"):
                        inside_code_block = True
                        code_block = []
                    elif line.strip().startswith("```") and inside_code_block:
                        inside_code_block = False
                        st.code("\n".join(code_block), language="python")
                    elif inside_code_block:
                        code_block.append(line)
                    else:
                        st.markdown(line)
                if "sandbox" in st.session_state:
                    qa_message = st.session_state["sandbox"].code_execution(self)
                    # FIX ME THIS IS UGLY
                    errors = list(qa_message.error_message.values())
                    errors = [error for error in errors if error is not None]
                    if len(errors) > 0:
                        # make the error message more visible with different color
                        st.markdown(f"Error: {qa_message.error_message}\n ")
                        # Remind users we are fixing the error by self debugging
                        st.markdown(f"Let me try to fix the error by self-debugging\n ")
                        if not debug:
                            st.session_state["sandbox"].llms["self_debug"].speak(
                                qa_message
                            )
                            qa_message = st.session_state["sandbox"].code_execution(
                                qa_message
                            )
                            self.render(debug=True)
                    # do not need to execute the block one more time
                    if not self.rendered:
                        self.rendered = True
                        st.session_state["sandbox"].render_qa_message(qa_message)

            elif render_key == "plots":
                # are there better ways in streamlit now to support plot display?
                plots = render_value
                for video_file_path, plot_list in plots.items():
                    for fig, axe in plot_list:
                        filename = save_figure_to_tempfile(fig)
                        st.image(filename, width=600)
            elif render_key == "out_videos":
                out_videos = render_value
                for video_path, videos in out_videos.items():
                    for video in videos:
                        if os.path.exists(video):
                            st.video(video)


class Messages:
    """
    The data class that is used in the front end (i.e., streamlit)
    methods
    -------
    parse_from_csv
        parse from example.csv for demo prompts
    render_message:
        render messages
    to_csv:
        to csv file

    attributes
    ----------
    {'plot': List[str],
    a list of file paths to image files that can be used by streamlit to render plots

    'code':  List[str],
    a list of str that contains functions. Though in the future,
    there might be more functions than the task program functions
    maybe it should be dictionary

    'ndarray': List[nd.array],
    Some sort of rendering for ndarray might be useful

    'text': List[str]
    The text needs to be further decomposed into different
    kinds of text, such as the main text, the plot documentation, the error
    each might have different colors and different positions
    }

    parameters
    ----------
    data: a dictionary
    """

    def __init__(self):
        self.raw_dict = None
        self.messages = []

    def parse_from_json(self, path):
        with open(path, "r") as f:
            json_obj = json.load(f)
        for json_entry in json_obj:
            if "query" in json_entry:
                self.append(HumanMessage(json_entry=json_entry))
            else:
                self.append(AIMessage(json_entry=json_entry))

    def render(self):
        """
        make sure those media types match what is in amadeus answer class
        """
        for message in self.messages:
            if isinstance(message, AIMessage):
                message.render()
            else:
                message.render()

    def append(self, e):
        self.messages.append(e)

    def insert(self, ind, e):
        self.messages.insert(ind, e)

    def __len__(self):
        return len(self.messages)

    def __iter__(self):
        return iter(self.messages)

    def __getitem__(self, ind):
        return self.messages[ind]

    def __setitem__(self, ind, value):
        self.messages[ind] = value


def get_amadeus_instance(example):
    # construct the config from the current example
    # get the root directory of the the module amadeusgpt
    identifier = get_identifier(example)
    amadeus_instance = AMADEUS(identifier.config)
    return amadeus_instance


def ask_amadeus(question):
    amadeus = get_amadeus_instance(st.session_state["example"])
    qa_message = amadeus.step(question)
    st.session_state["sandbox"] = amadeus.sandbox
    return qa_message


# caching display roi will make the roi stick to
# the display of initial state
def display_roi(analysis, example):
    analysis = create_analysis(get_identifier(example))
    roi_objects = analysis.get_roi_objects()

    frame = analysis.visual_manager.get_scene_image()
    colormap = plt.cm.get_cmap("rainbow", len(roi_objects))

    for i, roi_object in enumerate(roi_objects):
        name = roi_object.get_name()
        vertices = roi_object.Path.vertices
        pts = np.array(vertices, np.int32)
        pts = pts.reshape((-1, 1, 2))
        color = colormap(i)[:3]
        color = tuple(int(c * 255) for c in color[::-1])
        cv2.polylines(frame, [pts], isClosed=True, color=color, thickness=5)
        text = name
        font = cv2.FONT_HERSHEY_SIMPLEX
        text_position = (pts[0, 0, 0], pts[0, 0, 1])
        cv2.putText(frame, text, text_position, font, 1, color, 2, cv2.LINE_AA)
    with st.sidebar:
        st.caption("ROIs in the scene")
        st.image(frame)


def update_roi(analysis, result_json, ratios):
    w_ratio, h_ratio = ratios
    objects = pd.json_normalize(result_json["objects"])
    for col in objects.select_dtypes(include=["object"]).columns:
        objects[col] = objects[col].astype("str")
    roi_objects = {}
    objects = objects.to_dict(orient="dict")
    if "path" in objects:
        paths = objects["path"]
        count = 0
        for path_id in paths:
            temp = eval(paths[path_id])
            paths[path_id] = temp
            canvas_path = paths[path_id]
            if not isinstance(canvas_path, list):
                continue
            points = [[p[1], p[2]] for p in canvas_path if len(p) == 3]
            points = np.array(points)
            points[:, 0] = points[:, 0] * w_ratio
            points[:, 1] = points[:, 1] * h_ratio
            _object = ROIObject(f"ROI{count}", canvas_path=points)
            count += 1
            analysis.object_manager.add_roi_object(_object)

    analysis.object_manager.save_roi_objects("temp_roi_objects.pickle")


def place_st_canvas(analysis, key, scene_image):

    width, height = scene_image.size
    # we always resize the canvas to its default values and keep the ratio

    w_ratio = width / 600
    h_ratio = height / 400

    with st.sidebar:
        st.caption(
            "Left click to draw a polygon. Right click to confirm the drawing. Refresh the page if you need new ROIs or if the ROI canvas does not display"
        )
        canvas_result = st_canvas(
            fill_color="rgba(255, 165, 0, 0.9)",
            stroke_width=3,
            background_image=scene_image,
            # update_streamlit = realtime_update,
            width=600,
            height=400,
            drawing_mode="polygon",
            key=f"{key}_canvas",
        )

    if (
        canvas_result.json_data is not None
        and "path" in canvas_result.json_data
        and len(canvas_result.json_data["path"]) > 0
    ):
        pass
    if canvas_result.json_data is not None:
        update_roi(analysis, canvas_result.json_data, (w_ratio, h_ratio))

    if len(analysis.object_manager.get_roi_object_names()) > 0:
        display_roi(analysis, key)


def chat_box_submit():
    if "user_input" in st.session_state:
        query = st.session_state["user_input"]
        qa_message = ask_amadeus(query)

        user_message = HumanMessage(query=qa_message.query)
        amadeus_message = AIMessage(qa_message)

        st.session_state["messages"].append(user_message)
        st.session_state["messages"].append(amadeus_message)


def rerun_prompt(query, ind):
    messages = st.session_state["messages"]
    qa_message = ask_amadeus(query)
    amadeus_message = AIMessage(qa_message=qa_message)

    if ind != len(messages) - 1 and messages[ind + 1].role == "ai":
        messages[ind + 1] = amadeus_message
    else:
        messages.insert(ind + 1, amadeus_message)


def render_messages():
    example = st.session_state["example"]
    messages = st.session_state["messages"]

    if len(messages) == 0 and example != "Custom":
        example_history_json = os.path.join(f"examples/{example}/example.json")
        messages.parse_from_json(example_history_json)

    for ind, msg in enumerate(messages):
        _role = msg.role
        with st.chat_message(
            _role, avatar=USER_PROFILE if _role == "human" else BOT_PROFILE
        ):
            # parse and render message which is a list of dictionary
            if _role == "human":
                msg.render()
                disabled = not st.session_state["exist_valid_openai_api_key"]
                button_name = "Generate Response"
                st.button(
                    button_name,
                    key=f"{example}_user_{ind}",
                    on_click=rerun_prompt,
                    kwargs={"query": msg.query, "ind": ind},
                    disabled=disabled,
                )
            else:
                msg.render()

    st.session_state["messages"] = messages

    disabled = not st.session_state["exist_valid_openai_api_key"]
    st.chat_input(
        "Ask me new questions here ...",
        key="user_input",
        on_submit=chat_box_submit,
        disabled=disabled,
    )
    csv = None
    return csv


def update_df_data(new_item, index_to_update, df, csv_file):
    flat_items = [item[0] for item in new_item]
    df.iloc[index_to_update] = None
    df.loc[index_to_update, "Index"] = index_to_update
    for item in flat_items:
        key = item["type"]
        value = item["content"]
        if key in df.columns:
            df.loc[index_to_update, key] = value
    df.to_csv(csv_file, index=False)

    return csv_file, df


def get_scene_image(identifier: Identifier):
    analysis = create_analysis(identifier)

    scene_image = analysis.visual_manager.get_scene_image()
    if scene_image is not None:
        scene_image = Image.fromarray(scene_image)
        buffered = io.BytesIO()
        scene_image.save(buffered, format="JPEG")
        img_str = base64.b64encode(buffered.getvalue()).decode()
        return img_str


def get_identifier(example: str) -> Identifier:

    root_dir = os.path.dirname(amadeusgpt.__file__)
    if example == "Horse":
        template_config_path = os.path.join(root_dir, "configs", "Horse_template.yaml")
    elif example == "MausHaus":
        template_config_path = os.path.join(
            root_dir, "configs", "maushaus_template.yaml"
        )
    elif example == "EPM":
        template_config_path = os.path.join(root_dir, "configs", "EPM_template.yaml")
    elif example == "MABe":
        template_config_path = os.path.join(root_dir, "configs", "mabe_template.yaml")
    elif example == "Custom":
        template_config_path = os.path.join(root_dir, "configs", "Custom_template.yaml")

    config = Config(template_config_path)

    data_folder = os.path.join(Path(root_dir).parent, "examples", example)

    video_file_path = glob.glob(os.path.join(data_folder, "*.mp4"))[0]
    keypoint_file_path = glob.glob(os.path.join(data_folder, "*.h5"))[0]
    identifier = Identifier(config, video_file_path, keypoint_file_path)

    return identifier


def save_uploaded_file(uploaded_file, save_dir):

    filename = uploaded_file.name
    save_path = os.path.join(save_dir, filename)
    if not os.path.exists(save_path):
        with open(save_path, "wb") as f:
            f.write(uploaded_file.getbuffer())

    return save_path


def render_page_by_example(example):
    # get the config
    if example not in st.session_state:
        st.session_state[example] = {}
    # needs different logic for custom page
    identifier = get_identifier(example)
    config = identifier.config
    current_script_directory = os.path.dirname(os.path.abspath(__file__))
    logo_path = os.path.join(
        current_script_directory, "static", "images", "amadeusgpt_logo.png"
    )
    st.image(
        logo_path,
        caption=None,
        width=None,
        use_column_width=None,
        clamp=False,
        channels="RGB",
        output_format="auto",
    )
    # st.markdown("# Welcome to AmadeusGPT🎻")
    amadeus_root = Path(amadeusgpt.__file__).parent.parent

    if example == "Custom":
        st.markdown("Provide your own video and keypoint file (in pairs)")
        save_dir = os.path.join("examples", example)
        if "uploaded_keypoint_file" not in st.session_state:
            uploaded_keypoint_file = st.file_uploader(
                "Choose keypoint files",
                ["h5"],
                accept_multiple_files=False,
            )
            if uploaded_keypoint_file is not None:
                path = save_uploaded_file(uploaded_keypoint_file, save_dir)
                st.session_state["uploaded_keypoint_file"] = path
                config["data_info"]["data_folder"] = os.path.join("examples", example)
                config["data_info"]["data_folder"] = (
                    amadeus_root / config["data_info"]["data_folder"]
                )

        if "uploaded_video_file" not in st.session_state:
            uploaded_video_file = st.file_uploader(
                "Choose video files",
                VIDEO_EXTS,
                accept_multiple_files=False,
            )
            if uploaded_video_file is not None:
                path = save_uploaded_file(uploaded_video_file, save_dir)
                st.session_state["uploaded_video_file"] = uploaded_video_file
                config["data_info"]["data_folder"] = os.path.join("examples", example)
                config["data_info"]["data_folder"] = (
                    amadeus_root / config["data_info"]["data_folder"]
                )

        ###### USER INPUT PANEL ######
        # get user input once getting the uploaded files
        if uploaded_keypoint_file is None or uploaded_video_file is None:
            disabled = True
        else:
            disabled = False

        if disabled:
            st.warning("Please upload a file before entering text.")

    elif example == "EPM":
        st.markdown(
            "Elevated plus maze (EPM) is a widely used behavioral test. The mouse is put on an elevated platform with two open arms (without walls) and  two closed arms (with walls). \
                    In this example we used a video from https://www.nature.com/articles/s41386-020-0776-y."
        )
        st.markdown(
            "- ⬅️ On the left you can see the video data auto-tracked with DeepLabCut and keypoint names (below). You can also draw ROIs to ask questions to AmadeusGPT🎻 about the ROIs. You can drag the divider between the panels to increase the video/image size."
        )
        st.markdown(
            "- We suggest you start by clicking 'Generate Response' to our demo queries."
        )
        st.markdown(
            "- Ask additional questions in the chatbox at the bottom of the page."
        )
        st.markdown(
            "- Here are some example queries you might consider: 'The <|open arm|> is the ROI0. How much time does the mouse spend in the open arm?' (NOTE here you can re-draw an ROI0 if you want. Be sure to click 'finish drawing') | 'Define head_dips as a behavior where the mouse's mouse_center and neck are in ROI0 which is open arm while head_midpoint is outside ROI1 which is the cross-shape area. When does head_dips happen and what is the number of bouts for head_dips?' "
        )
        st.markdown("- ⬇️🎥 Watch this short clip on how to draw the ROI(s)🤗")

        st.video(
            os.path.join(current_script_directory, "static/customEPMprompt_short.mp4")
        )

    elif example == "MABe":
        st.markdown(
            "MABe Mouse Triplets is part of a behavior benchmark presented in Sun et al 2022 https://arxiv.org/abs/2207.10553. In the videos, three mice exhibit multiple social behaviors including chasing.   \
                    In this example, we take one video where chasing happens between mice."
        )
        st.markdown(
            "- ⬅️ On the left you can see the video data and keypoint names, which could be useful for your queries. You can drag the divider between the panels to increase the video/image size."
        )
        st.markdown(
            "- We suggest you start by clicking 'Generate Response' to our demo queries."
        )
        st.markdown(
            "- Ask additional questions in the chatbox at the bottom of the page."
        )
        analysis = create_analysis(identifier)

    elif example == "MausHaus":
        st.markdown(
            "MausHaus is a dataset that records a freely moving mouse within a rich environment with objects. More details can be found https://arxiv.org/pdf/2203.07436.pdf."
        )
        st.markdown(
            "- ⬅️ On the left you can see the video data auto-tracked with DeepLabCut, segmented image with SAM, and the keypoint guide, which could be useful for your queries. You can drag the divider between the panels to increase the video/image size."
        )
        st.markdown(
            "- We suggest you start by clicking 'Generate Response' to our demo queries."
        )
        st.markdown(
            "- Ask additional questions in the chatbox at the bottom of the page."
        )
        st.markdown(
            "- Here are some example queries you might consider: 'Give me events where the animal overlaps with the treadmill, which is object 5' | 'Define <|drinking|> as a behavior where the animal's nose is over object 28, which is a waterbasin. The minimum time window for this behavior should be 20 frames. When is the animal drinking?'"
        )

    elif example == "Horse":
        st.markdown(
            "This horse video is part of a benchmark by Mathis et al 2021 https://arxiv.org/abs/1909.11229."
        )

    identifier = get_identifier(example)

    analysis = create_analysis(identifier)

    if st.session_state["example"] != example:
        st.session_state["messages"] = Messages()
    st.session_state["example"] = example

    scene_image_path = get_scene_image(identifier)

    if scene_image_path is not None:
        img_data = base64.b64decode(scene_image_path)
        image_stream = io.BytesIO(img_data)
        image_stream.seek(0)
        scene_image = Image.open(image_stream)
    else:
        scene_image = None

    col1, col2, col3 = st.columns([2, 1, 1])

    with st.sidebar as sb:
        if example == "MABe":
            st.caption("Raw video from MABe")
        elif example == "Horse":
            st.caption("Raw video from Horse-30")
        else:
            st.caption("DeepLabCut-SuperAnimal tracked video")

        if os.path.exists(identifier.video_file_path):
            st.video(identifier.video_file_path)

        if "uploaded_video_file" in st.session_state:
            st.video(st.session_state["uploaded_video_file"])

    if (
        st.session_state["example"] == "EPM"
        or st.session_state["example"] == "MausHaus"
        and scene_image is not None
    ):
        place_st_canvas(analysis, example, scene_image)

    if st.session_state["example"] == "Custom" and scene_image:
        place_st_canvas(analysis, example, scene_image)

    if example == "EPM" or example == "MausHaus":
        # will read the keypoints from h5 file to avoid hard coding
        with st.sidebar:
            topviewimage = os.path.join(
                current_script_directory, "static", "images", "supertopview.png"
            )
            st.image(topviewimage)
            # st.image("static/images/supertopview.png")
    with st.sidebar:
        st.write("Keypoints:")
        st.write(analysis.get_keypoint_names())

    render_messages()
    gc.collect()


def save_figure_to_tempfile(fig):
    # save the figure
    folder_path = os.path.join("logs", "tmp_imgs")
    if not os.path.exists(folder_path):
        os.makedirs(folder_path)
    # Generate a unique temporary filename in the specified folder
    temp_file = tempfile.NamedTemporaryFile(
        dir=folder_path, suffix=".png", delete=False
    )
    filename = temp_file.name
    temp_file.close()
    fig.savefig(
        filename,
        format="png",
        bbox_inches="tight",
        pad_inches=0.0,
        dpi=400,
        transparent=True,
    )
    return filename


def make_plot_pretty4dark_mode(fig, ax):
    fig = plt.gcf()
    fig.set_facecolor("none")
    ax = plt.gca()
    ax.set_facecolor("none")
    # Set axes and legend colors to white or other light colors
    ax.spines["bottom"].set_color("white")
    ax.spines["top"].set_color("white")
    ax.spines["right"].set_color("white")
    ax.spines["left"].set_color("white")

    ax.xaxis.label.set_color("white")
    ax.yaxis.label.set_color("white")
    ax.title.set_color("white")
    ax.tick_params(axis="x", colors="white")
    ax.tick_params(axis="y", colors="white")
    legend = plt.legend()
    for text in legend.get_texts():
        text.set_color("white")

    return fig, ax


def display_image(temp_file):
    full_image = Image.open(temp_file)
    st.image(full_image)


def display_temp_text(text_content):
    # Convert the text content to base64
    text_bytes = text_content.encode("utf-8")
    text_base64 = base64.b64encode(text_bytes).decode()
    # Display the link to the text file
    st.markdown(
        f'<a href="data:text/plain;charset=utf-8;base64,{text_base64}" target="_blank">Check error.</a>',
        unsafe_allow_html=True,
    )


def style_button_row(clicked_button_ix, n_buttons):
    def get_button_indices(button_ix):
        return {"nth_child": button_ix, "nth_last_child": n_buttons - button_ix + 1}

    clicked_style = """
    div[data-testid*="stHorizontalBlock"] > div:nth-child(%(nth_child)s):nth-last-child(%(nth_last_child)s) button {
        border-color: rgb(255, 75, 75);
        color: rgb(255, 75, 75);
        box-shadow: rgba(255, 75, 75, 0.5) 0px 0px 0px 0.2rem;
        outline: currentcolor none medium;
    }
    """
    unclicked_style = """
    div[data-testid*="stHorizontalBlock"] > div:nth-child(%(nth_child)s):nth-last-child(%(nth_last_child)s) button {
        pointer-events: none;
        cursor: not-allowed;
        opacity: 0.65;
        filter: alpha(opacity=65);
        -webkit-box-shadow: none;
        box-shadow: none;
    }
    """
    style = ""
    for ix in range(n_buttons):
        ix += 1
        if ix == clicked_button_ix:
            style += clicked_style % get_button_indices(ix)
        else:
            style += unclicked_style % get_button_indices(ix)
    st.markdown(f"<style>{style}</style>", unsafe_allow_html=True)



import json
import os
from abc import ABC, abstractmethod
from functools import wraps

from cachetools import LRUCache

from amadeusgpt.config import Config


class BaseManager(ABC):
    """
    The subclass of this should maintain lists of serializeable objects.
    During serialization, it should serialize all the objects and save them to a json file.
    During deserialization, it should read the json file and deserialize all the objects.
    """

    @abstractmethod
    def serialize(self, base_path: str):
        pass

    @abstractmethod
    def deserialize(self, base_path: str):
        pass


def make_hashable(obj):
    if isinstance(obj, (tuple, list)):
        return tuple(make_hashable(e) for e in obj)
    elif isinstance(obj, dict):
        return frozenset((k, make_hashable(v)) for k, v in obj.items())
    elif isinstance(obj, set):
        # Convert set to frozenset, which is hashable
        return frozenset(make_hashable(e) for e in obj)
    else:
        return obj


class cache_decorator:
    def __init__(self, func):
        self.func = func
        wraps(func)(self)

    def __get__(self, instance, owner):
        if instance is None:
            return self
        # Return a new, bound version of the decorator, with the instance bound.

        return lambda *args, **kwargs: self.__call__(instance, *args, **kwargs)

    def __call__(self, *args, **kwargs):
        # The first argument is now the instance.
        instance, *args = args

        if not instance.use_cache:
            return self.func(instance, *args, **kwargs)

        hashable_args = make_hashable(args)

        hashable_kwargs = make_hashable(frozenset(kwargs.items()))

        cache_key = (hashable_args, hashable_kwargs)

        if cache_key in instance._cache:
            return instance._cache[cache_key]

        result = self.func(instance, *args, **kwargs)
        instance._cache[cache_key] = result
        return result


class Manager(BaseManager):
    def __init__(self, config: Config | dict, use_cache: bool = False):
        self.config = config
        self.use_cache = use_cache
        self._cache = LRUCache(maxsize=128)

    def serialize(self, base_path: str):
        ret = {}

        for attr_name in self.__dict__:
            if attr_name in self.get_serializeable_list_names():
                object_list = self.__dict__[attr_name]
                for idx, object in enumerate(object_list):
                    ret[attr_name].append(object.serialize(base_path, idx))

            else:
                ret[attr_name] = self.__dict__[attr_name]

        with open(os.path.join(base_path, f"{self.__class__.__name__}.json"), "w") as f:
            json.dump(ret, f)

    def deserialize(self, base_path: str) -> BaseManager:
        json_path = os.path.join(base_path, f"{self.__class__.__name__}.json")
        with open(json_path, "r") as f:
            data = json.load(f)

        serializeable_list_names = self.get_serializeable_list_names()

        for attr_name in data:
            if attr_name in serializeable_list_names:
                serialized_data = data[attr_name]
                for idx, serialized_data in serialized_data.items():
                    instance = object.deserialize(serialized_data)
                    self.__dict__[attr_name].append(instance)
            else:
                self.__dict__[attr_name] = data[attr_name]

    def summary(self):
        for attr_name in self.__dict__:
            if attr_name in self.get_serializeable_list_names():
                print(f"{attr_name} has {len(self.__dict__[attr_name])} objects")
                if len(self.__dict__[attr_name]) > 0:
                    self.__dict__[attr_name][0].summary()
            else:
                print(f"{attr_name} has {self.__dict__[attr_name]}")



# from google.cloud import firestore
import logging
import os
import platform
from collections import Counter
from datetime import datetime

import streamlit as st


class AmadeusLogger:
    if "streamlit_cloud" in os.environ:
        level = logging.DEBUG
    else:
        level = logging.INFO
    logging.basicConfig(level=level)
    filename = None
    logger = logging.getLogger()
    if "streamlit_cloud" in os.environ:
        db = firestore.Client.from_service_account_json(
            "amadeusgpt-a1849-firebase-adminsdk-eyfx1-4a1bbc8488.json"
        )

    @classmethod
    def format(cls, message):
        session_id = st.session_state.get("session_id", "")
        now = datetime.now()
        timestamp = now.strftime("%Y%m%d_%H%M%S")
        username = st.session_state["username"] = "no_username"
        email = st.session_state["email"] = "no_email"
        return f"{timestamp} - session_id:{session_id} - username:{username} - email: {email} - {message}"

    @classmethod
    def info(cls, msg):
        cls.logger.info(cls.format(msg))

    @classmethod
    def debug(cls, msg):
        cls.logger.debug(cls.format(msg))

    @classmethod
    def store_chats(cls, key, value):
        if "streamlit_cloud" in os.environ:
            if "username" in st.session_state:
                username = st.session_state["username"]
                doc_ref = cls.db.collection("chats").document(username)
                doc = doc_ref.get()
                if doc.exists:
                    _dict = doc.to_dict()
                    if key not in _dict:
                        _dict[key] = []
                    _dict[key].append(cls.format(value))
                    doc_ref.update(_dict)
                else:
                    doc_ref.set({key: [cls.format(value)]})

        else:
            # if not running in the cloud:
            cls.logger.info(cls.format(value))

    @classmethod
    def log_process_memory(cls, log_position=""):
        """
        We use the memory consumption from the perspective of every user, right after ask amadeus
        """
        process = psutil.Process(os.getpid())
        # Print the current memory usage
        memory_str = f"Current memory usage: {process.memory_info().rss / 1024**2} MB"
        cls.debug(cls.format(f"{log_position}: {memory_str}"))
        if "streamlit_cloud" in os.environ:
            if "username" in st.session_state:
                username = st.session_state["username"]
                doc_ref = cls.db.collection("process_memory").document(username)
                doc = doc_ref.get()
                now = datetime.now()
                timestamp = now.strftime("%Y%m%d_%H%M%S")
                temp = {
                    "memory_profile": [
                        {
                            "timestamp": timestamp,
                            "memory_consumption": memory_str,
                            "operating_system": platform.system(),
                            "streamlit_app": os.environ["streamlit_app"],
                            "streamlit_cloud": os.environ["streamlit_cloud"],
                            "meta": cls.format(""),
                            "log_position": log_position,
                        }
                    ]
                }
                if doc.exists:
                    _dict = doc.to_dict()
                    _dict["memory_profile"].extend(temp["memory_profile"])
                    doc_ref.update(_dict)
                else:
                    doc_ref.set(temp)

    @classmethod
    def check_user_exist(cls, username):
        if "streamlit_cloud" in os.environ:
            doc_ref = cls.db.collection("chats").document(username)
            doc = doc_ref.get()
            return doc.exists
        else:
            return False

    @classmethod
    def log_feedback(cls, feedback_type, content):
        if "streamlit_cloud" in os.environ:
            user_msg, bot_msg = content
            content = cls.format(str(user_msg) + str(bot_msg))
            username = st.session_state.get("username", "fake_username")
            doc_ref = cls.db.collection("feedbacks").document(username)
            temp = {"content": content, "feedback_type": feedback_type}
            doc = doc_ref.get()
            if doc.exists:
                _dict = doc.to_dict()
                _dict["feedback"].append(temp)
                doc_ref.update(_dict)
            else:
                doc_ref.set({"feedback": [temp]})

    @classmethod
    def log(cls, message, level=1):
        if "streamlit_cloud" in os.environ:
            if "username" in st.session_state:
                username = st.session_state["username"]
                doc_ref = cls.db.collection("chats").document(username)
                doc = doc_ref.get()
                if doc.exists:
                    _dict = doc.to_dict()
                    if f"log_level{level}" not in _dict:
                        _dict[f"log_level{level}"] = []
                    _dict[f"log_level{level}"].append(message)
                    doc_ref.update(_dict)
                else:
                    doc_ref.set({f"log_level{level}": [message]})
        else:
            # if not running in the cloud:
            cls.logger.info(cls.format(message))


def parse_logs(db):
    chats_coll = db.collection("chats")
    fbs_coll = db.collection("feedbacks")

    n_rephrasing = 0
    queries = []
    errors = []
    for chat in chats_coll.stream():
        dict_ = chat.to_dict()
        if "log_level1" in dict_:
            log = [s if s else "" for s in dict_["log_level1"]]
            queries_ = set()
            for l in log:
                if "Before rephrasing: " in l:
                    queries_.add(l.replace("Before rephrasing: ", ""))
            n_rephrasing += len(queries_)
        elif "user_query" in dict_:
            queries_ = set([q.split(" - ")[-1] for q in dict_["user_query"]])
        else:
            raise ValueError("Ups")

        queries.extend(queries_)
        if "code_generation_errors" in dict_:
            errors_ = set([e.split(" - ")[-1] for e in dict_["code_generation_errors"]])
            errors.extend(errors_)
    error_messages = []
    for error in errors:
        if "query" in error:
            e = error.replace("'", "").split("error_message: ")[1].split("Traceback")[0]
            error_messages.append(e)

    n_queries = len(queries)
    n_errors = len(errors)

    fbs = []
    for fb in fbs_coll.stream():
        if fb.id not in ("like", "dislike"):
            fbs.append(fb.to_dict())
    fb_types = [f["feedback_type"] for fb in fbs for f in fb["feedback"]]
    fb_types.extend(
        ["like"] * len(fbs_coll.document("like").get().to_dict()["feedback"])
    )
    fb_types.extend(
        ["dislike"] * len(fbs_coll.document("dislike").get().to_dict()["feedback"])
    )
    counter = Counter(fb_types)

    return {
        "n_queries": n_queries,
        "n_errors": n_errors,
        "queries": queries,
        "error_messages": error_messages,
        "n_rephrasing": n_rephrasing,
        "n_likes": counter["like"],
        "n_dislikes": counter["dislike"],
    }



import os

from amadeusgpt.config import Config


class Identifier:
    """
    config contains project level meta
    other parameters help identify the unique params needed for animal behavior analysis
    Currently video_file_path and keypoint_file_path.
    Can be more in the future
    """

    def __init__(
        self, config: Config | dict, video_file_path: str, keypoint_file_path: str
    ):

        self.config = config
        self.video_file_path = video_file_path
        self.keypoint_file_path = keypoint_file_path

    def __str__(self):
        return f"""------
video_file_path: {self.video_file_path} 
keypoint_file_path: {self.keypoint_file_path}
config: {self.config}
------
"""

    def __eq__(self, other):
        if os.path.exists(self.video_file_path):
            return os.path.abspath(self.video_file_path) == os.path.abspath(
                other.video_file_path
            )
        else:
            return os.path.abspath(self.keypoint_file_path) == os.path.abspath(
                other.keypoint_file_path
            )

    def __hash__(self):
        if os.path.exists(self.video_file_path):
            return hash(os.path.abspath(self.video_file_path))
        else:
            return hash(os.path.abspath(self.keypoint_file_path))



from typing import Any, Dict, List

import cv2
import matplotlib.lines as mlines
import matplotlib.pyplot as plt
from matplotlib.path import Path
from matplotlib.widgets import PolygonSelector

from amadeusgpt.analysis_objects.object import ROIObject
from amadeusgpt.behavior_analysis.identifier import Identifier
from amadeusgpt.programs.api_registry import register_class_methods

from .base import Manager
from .object_manager import ObjectManager


class ROISelector:
    roi_count = 0
    cmap = ["red", "blue", "yellow", "green"]

    def __init__(self, axs, object_manager):
        self.axs = axs
        self.object_manager = object_manager
        self.selector = PolygonSelector(self.axs, self.onselect)
        self.paths = []

    def roi_select_event(self, vertices):
        # once the bounding box is done drawing, run the following command
        first_point = vertices[0]
        vertices.append(first_point)
        path = Path(vertices)
        self.paths.append(path)
        # self.axs.clear()
        for i, path in enumerate(self.paths):
            self.axs.add_patch(
                plt.Polygon(path.vertices, fill=None, edgecolor=type(self).cmap[i])
            )
        handles = [
            mlines.Line2D([], [], color=self.cmap[i], label=f"ROI{i}")
            for i in range(len(self.paths))
        ]
        self.axs.legend(handles=handles, loc="upper right")

        # saving roi figure

    def onselect(self, vertices):
        self.roi_select_event(vertices)
        figure_output = "roi_figure.png"
        plt.savefig(figure_output, dpi=800)

        # Here you can add any further processing of the polygons
        self.object_manager.roi_objects = []
        for idx, path in enumerate(self.paths):
            self.object_manager.add_roi_object(ROIObject(f"ROI{idx}", path))

        # Assuming the object_manager's add_roi_object is meant to handle the completed polygons


@register_class_methods
class GUIManager(Manager):
    def __init__(self, identifier: Identifier, object_manager: ObjectManager):
        self.config = identifier.config
        self.video_file_path = identifier.video_file_path
        self.object_manager = object_manager
        if self.video_file_path is None:
            return
        self.videos = {}

    def add_roi_from_video_selection(self):
        cap = cv2.VideoCapture(self.video_file_path)
        if not cap.isOpened():
            print("Error opening video file.")
            return []

        total_frames = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
        middle_frame_index = total_frames // 2
        cap.set(cv2.CAP_PROP_POS_FRAMES, middle_frame_index)
        ret, frame = cap.read()
        cap.release()

        if not ret:
            print("Failed to retrieve frame.")
            return []
        fig, ax = plt.subplots(1)
        ax.imshow(frame)
        self.selector = ROISelector(ax, self.object_manager)

    def get_serializeable_list_names(self) -> List[str]:
        return []



import os
import pprint

import yaml


def create_project(data_folder, result_folder, **kwargs):
    """
    Create a project config file. Save the config file to the result folder
    """
    config = {
        "data_info": {
            "data_folder": data_folder,
            "result_folder": result_folder,
            "video_suffix": ".mp4",
        },
        "llm_info": {"max_tokens": 4096, "temperature": 0.0, "keep_last_n_messages": 2},
        "object_info": {"load_objects_from_disk": False, "use_grid_objects": False},
        "keypoint_info": {
            "use_3d": False,
            "include_confidence": False,
        },
        "video_info": {"scene_frame_number": 1},
    }
    # save the dictionary config to yaml

    def set_nested_value(d, keys, value):
        for key in keys[:-1]:
            d = d.setdefault(key, {})
        d[keys[-1]] = value

    for key, value in kwargs.items():
        keys = key.split(".")
        set_nested_value(config, keys, value)

    os.makedirs(result_folder, exist_ok=True)

    file_path = os.path.join(result_folder, "config.yaml")

    with open(file_path, "w") as f:
        yaml.dump(config, f)

    print(
        f"Project created at {result_folder}. Results will be saved to {result_folder}"
    )
    print(
        f"The project will load video files (*{config['data_info']['video_suffix']}) and optionally keypoint files from {data_folder}"
    )
    print(f"A copy of the project config file is saved at {file_path}")
    pprint.pprint(config)

    return config



def _get_system_prompt(user_input, thought_process, answer):
    system_prompt = f""" You are an expert on animal behavior analysis. Your job is to interpret the answer we give to our users who are neuroscinetist that are asking questinos about animal behaviors.
    The questions was {user_input}
    The thought process for the answer was {thought_process}
    The answer was {answer}
    Now explain to the users (who asked questions) the answer. Make sure your explanation is only about high level concepts of animal behavior analysis. Do not mention anything about
    the function or the code.
    """
    return system_prompt



from .animal_manager import AnimalManager
from .base import Manager
from .event_manager import EventManager
from .gui_manager import GUIManager
from .object_manager import ObjectManager
from .relationship_manager import RelationshipManager
from .visual_manager import VisualManager



from typing import List

import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.collections import LineCollection
from scipy import signal

from amadeusgpt.programs.api_registry import register_integration_api


def find_peaks(x, min_dist=50, detrend=False):
    if detrend:
        x = signal.detrend(x[~np.isnan(x)])
    return signal.find_peaks(x, distance=min_dist)[0]


def next_pow_two(val):
    return 1 << (val - 1).bit_length()


def autocorr(x, axis=-1):
    """Compute the autocorrelation of a 1D time series."""
    x = np.atleast_1d(x)
    length = len(x)
    n = next_pow_two(length)
    f = np.fft.fft(x - np.mean(x, axis=axis), n=2 * n, axis=axis)
    acf = np.fft.ifft(f * np.conjugate(f), axis=axis)[:length].real
    acf /= acf[0]
    return acf


def get_events(analysis, event, kpt_names, min_dist=None, detrend=False):

    if event not in ("contact", "lift"):
        raise ValueError('`event` must be either "contact" or "lift".')

    coords = analysis.get_keypoints().copy()
    # center the keypoints
    coords -= np.nanmedian(coords, axis=2)[:, None]
    events = []
    for kpt in kpt_names:
        x = coords[:, 0, analysis.get_keypoint_names().index(kpt), 0]
        if event == "lift":
            x = -x
        if min_dist is None:  # Determine a signal's period from its autocorrelation
            corr = autocorr(x[~np.isnan(x)])
            start = np.argmin(corr)
            period = np.argmax(corr[start:]) + start
            min_dist = int(0.75 * period)  # To be safe
        peaks = find_peaks(x, min_dist, detrend)
        events.append(peaks)
    return events


def calc_stride_durations(contacts):
    return [np.diff(c) for c in contacts]


def calc_stride_lengths(analysis, keypoints, hoof_kpt_names, contacts):
    stride_lengths = []
    for hoof, contacts_ in zip(hoof_kpt_names, contacts):
        x = keypoints[
            contacts_, :, analysis.get_keypoint_names().index(hoof), 0
        ].flatten()
        stride_lengths.append(np.diff(x))
    return stride_lengths


def calc_duty_factors(contacts, lifts):
    duty = []
    for contacts_, lifts_ in zip(contacts, lifts):
        temp = []
        for c1, c2 in zip(contacts_[:-1], contacts_[1:]):  # Individual strides
            lift = lifts_[(lifts_ > c1) & (lifts_ < c2)]
            if not lift.size:
                continue
            d = (lift[0] - c1) / (c2 - c1)
            temp.append(d)
        duty.append(temp)
    return duty


def get_stances(contacts, lifts):
    stances = []
    for contacts_, lifts_ in zip(contacts, lifts):
        temp = set()
        for c1, c2 in zip(contacts_[:-1], contacts_[1:]):  # Individual strides
            lift = lifts_[(lifts_ > c1) & (lifts_ < c2)]
            if not lift.size:
                continue
            temp.add((c1, lift[0]))
        for l1, l2 in zip(lifts_[:-1], lifts_[1:]):
            contact = contacts_[(contacts_ > l1) & (contacts_ < l2)]
            if not contact.size:
                continue
            temp.add((contact[0], l2))
        stances.append(list(temp))
    return stances


@register_integration_api
def run_gait_analysis(self, limb_keypoint_names: List[str]) -> dict:
    """
    This function computse an animal's gait parameters given a list of distal keypoints.
    Parameters
    ----------
    limb_keypoint_names: List[str], list of the names of the distal keypoints. Need to be at least 2 keypoints.
    """
    min_dist = None
    contacts = get_events(self, "contact", limb_keypoint_names, min_dist)
    lifts = get_events(self, "lift", limb_keypoint_names, min_dist)
    stride_durations = calc_stride_durations(contacts)
    keypoints = self.get_keypoints()
    stride_lengths = calc_stride_lengths(self, keypoints, limb_keypoint_names, contacts)
    duty_factors = calc_duty_factors(contacts, lifts)
    stances = get_stances(contacts, lifts)
    return {
        "contacts": contacts,
        "lifts": lifts,
        "stride_durations": stride_durations,
        "stride_lengths": stride_lengths,
        "duty_factors": duty_factors,
        "stances": stances,
    }


def _make_line_collection(
    coords, links, start=0, end=-1, inds=None, color_stance="plum", alpha=0.5
):
    color = mcolors.to_rgb("gray")
    colors = np.array([color] * len(coords))
    if inds is not None:
        mask = np.zeros(coords.shape[0], dtype=bool)
        for ind1, ind2 in inds:
            mask[ind1 : ind2 + 1] = True
        colors[mask] = mcolors.to_rgb(color_stance)
    sl = slice(start, end)
    colors = colors[sl]

    segs = coords[sl, links].reshape((-1, 2, 2))
    colors = np.repeat(colors, len(links), axis=0)
    coll = LineCollection(segs, colors=colors, alpha=alpha)
    return coll, segs


@register_integration_api
def plot_gait_analysis_results(
    self, gait_analysis_results, limb_keypoints, color_stance="plum"
):
    """
    This function plots the gait analysis results returned from the `run_gait_analysis` function.
    Parameters
    ----------
    gait_analysis_results: dict, the results from the `run_gait_analysis` function
    limb_keypoints: List[str], list of the names of the distal keypoints. Need to be at least 2 keypoints.
    color_stance: str, optional, default to be "plum"
    """
    fig, ax = plt.subplots(sharex=True, sharey=True)
    coords = self.get_keypoints()[:, 0]

    if gait_analysis_results["stances"] == [[]]:
        return fig, ax
    stance_inds = gait_analysis_results["stances"][0]

    for spine in ax.spines.values():
        spine.set_visible(False)
    ax.set_xticks([])
    ax.set_yticks([])
    ax.set_aspect("equal")

    skeleton = []
    for kpt1, kpt2 in zip(limb_keypoints, limb_keypoints[1:]):
        skeleton.append(
            (
                self.get_keypoint_names().index(kpt1),
                self.get_keypoint_names().index(kpt2),
            ),
        )
    coll, segs = _make_line_collection(
        coords, skeleton, inds=stance_inds, color_stance=color_stance
    )
    xmin, ymin = np.nanmin(segs, axis=(0, 1))
    xmax, ymax = np.nanmax(segs, axis=(0, 1))
    ax.set_xlim(xmin, xmax)
    ax.set_ylim(ymin, ymax)
    ax.invert_yaxis()
    ax.add_collection(coll)
    ax.set_ylabel("Limb")
    return fig, ax
